#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include "debug.h"

#define NAP_IN		pipe_in[0]
#define NAP_OUT		pipe_out[1]

pthread_attr_t thread_attr;
int pipe_in[2], pipe_out[2];
int nap_pid;
char *progname;
int debuglevel;
pthread_t read_thread;

static void *read_process (void *dummy)
{
  char c;

  log_debug (255, "read_process: starting");

  while (read (NAP_IN, &c, 1) > 0)
    log_debug (255, "read_process: %c", c);

  log_debug (255, "read_process: exiting");
  pthread_exit (0);
}

static void got_interrupt (int sig)
{	
  log_debug (DEBUG_NAPSTER, "Got interrupt %d!", sig);
  if (nap_pid)
    kill (SIGTERM, nap_pid);

  close_debug ();
  exit (1);
}

static void my_sleep (int t)
{
  log_debug (DEBUG_NAPSTER, "my_sleep: %d", t);
  sleep (t);
}

static void send_nap (char *s)
{
  log_debug (DEBUG_NAPSTER, "send_nap: %s", s);
  write (NAP_OUT, s, strlen(s));
}


main (int ac, char **av)
{
  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  init_debug ();

  if (pipe (pipe_in) < 0) {
    log_err ("pipe_in: %s", strerror(errno));
    exit (1);
  }

  if (pipe (pipe_out) < 0) {
    log_err ("pipe_out: %s", strerror(errno));
    exit (1);
  }

  if (pthread_create (&read_thread, &thread_attr, read_process, NULL) != 0)
    log_err ("pthread_create error for read_process");

  if (!(nap_pid = fork ())) {
    dup2 (pipe_out[0], 0);
    close (pipe_out[0]);

    if (execlp (av[1], av[1], NULL) < 0) {
      log_err ("execlp: %s: %s", av[1], strerror(errno));
      exit (1);
    }
  }
  else {
    signal (SIGINT, got_interrupt);
    signal (SIGQUIT, got_interrupt);
    signal (SIGTERM, got_interrupt);

    my_sleep (20);
    send_nap ("/search souchon\r");
    my_sleep (20);
    send_nap ("j");
    my_sleep (2);
    send_nap ("j");
    my_sleep (10);
    send_nap ("q\r/quit\r");

    exit (0);

  }
}
