#ident "$Id: cdplayer.c,v 1.6 2000/12/20 07:25:31 pierre Exp pierre $"
/*
 * OMM CD player
 */

#define EXTERN

#include "omm.h"
#include "cdplayer.h"

static struct disc_info disc_info;
static struct disc_data disc_data;
static int fd_cd;
static int cd_check_count;
static int first_audio_track;
static int current_track;
static pthread_t check_cd_thread;
static int old_minutes, old_seconds, old_current_track;
static int display_type = DISPLAY_CD_LENGTH;
static int old_display_type = -1;
static int old_disc_mode;
static int old_disc_present = -1;
static boolean use_stdin = FALSE;
static boolean tray_closed;
static int disc_id;                /* Current disc Id */
static char *omm_dir = DEFAULT_OMMDIR;
static char disc_spool_dir[MAXPATHLEN];

/* States handling definition */
static omm_event_handler cd_event_handlers[] = {
  'd', OMM_STATE_INIT, lcd_dump_screen,		/* Dump screen to log file */
  'd', STATE_CD, lcd_dump_screen,
  'B', STATE_CD, lcd_bg_screen,
  'B', OMM_STATE_INIT, lcd_bg_screen,
  'F', STATE_CD, lcd_fg_screen,
  'F', OMM_STATE_INIT, lcd_fg_screen,
  'e', STATE_CD, cd_player_eject_close,		/* Open/close tray */
  'e', OMM_STATE_INIT, cd_player_eject_close,	/* Open/close tray */
  'p', STATE_CD, cd_player_play,		/* Play */
  'p', OMM_STATE_INIT, cd_player_play,		/* Play */
  's', STATE_CD, cd_player_stop,		/* Stop */
  'u', STATE_CD, cd_player_mark_unmark_track,	/* Mark/Unmark track */
  'U', STATE_CD, cd_player_mark_unmark_all,	/* Mark/Unmark all tracks*/
  'h', STATE_CD, cd_player_pause,		/* Pause */
  '+', STATE_CD, cd_player_next_track,		/* Track + */
  '-', STATE_CD, cd_player_previous_track,	/* Track - */
  'q', STATE_CD, cd_player_quit,		/* Quit */
  'q', OMM_STATE_INIT, cd_player_quit,		/* Quit */
  't', STATE_CD, cd_player_toggle_display,	/* Toggle display: time/title */
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, cd_player_default_handler,
  0, 0, NULL
};

static char *mode_msg[] = { "PLAY", "PAUSE", "STOP", "STOP", "INVALID", "ERROR" };

static int explore_dir (char *dir)
{
  DIR *dirfile;
  struct dirent *anentry;
  struct stat statb;
  int count = 0;

  chdir (dir);
  dirfile = opendir(".");

  while ((anentry = readdir (dirfile)) != NULL) {
    /* skip some files */
    if (anentry->d_name[0] == '.' || stat(anentry->d_name, &statb) < 0) 
      continue;
    else if (S_ISDIR (statb.st_mode))
      return explore_dir (anentry->d_name);
    else
      count++;
  }

  closedir (dirfile);

  return count;
}

static void remove_spool_dir ()
{
  char buf[MAXPATHLEN];

  /* Try to remove .disc_info + spool dir */
  if (!explore_dir (disc_spool_dir)) {
    sprintf (buf, "%s/%s", disc_spool_dir, DISC_INFO_FILE);
    unlink (buf);
    rmdir (disc_spool_dir);
  }
}

/*
 * Display functions
 */

static void display_no_disc ()
{
  lcd_clear_screen ();
  lcd_write_center (2, "NO DISC!", ATTR_NORMAL);
}

static void display_no_audio ()
{
  lcd_clear_screen ();
  lcd_write_center (2, "NO AUDIO TRACK!", ATTR_NORMAL);
}

static void display_zero_time ()
{
  char buf[LCD_WIDTH+1];

  lcd_init_hscroll ();
  sprintf (buf, "Track   00:00 %02d", current_track);
  lcd_write_center (1, buf, ATTR_NORMAL);
}

static void display_track_length ()
{
  char buf[LCD_WIDTH+1];

  if (display_type != DISPLAY_TRACK_LENGTH)
    return;

  if (current_track > 0) {
    sprintf (buf, "Length   %02d:%02d   ", disc_info.disc_track[current_track-1].track_length.minutes, disc_info.disc_track[current_track-1].track_length.seconds, current_track);
    lcd_clear_line (2);
    lcd_write_center (2, buf, ATTR_NORMAL);
  }
}

static void display_disc_length ()
{
  char buf[LCD_WIDTH+1];

  if (display_type != DISPLAY_CD_LENGTH)
    return;

  sprintf (buf, "Total   %02d:%02d %02d", disc_info.disc_length.minutes, disc_info.disc_length.seconds, disc_info.disc_total_tracks);
  lcd_write_center (2, buf, ATTR_NORMAL);
}

static void display_disc_info (struct disc_info *info)
{
  char buf[LCD_WIDTH+1];

  if (!first_audio_track)
    return;

  /* New disc inserted: display general info */
  if (old_disc_present != info->disc_present) {
    lcd_clear_screen ();
    lcd_write (0, 0, "CDA", ATTR_NORMAL);
    current_track = first_audio_track;
    display_zero_time ();
    display_track_length ();
    display_disc_length ();
    display_track_selection ();
  }
  else {
    if (old_disc_mode != info->disc_mode) {
      /* Special case of end-of-disc */
      if ((info->disc_mode == CDAUDIO_NOSTATUS || info->disc_mode == CDAUDIO_COMPLETED) && (old_disc_mode == CDAUDIO_PLAYING)) {
	cd_player_stop ();
      }

      /* Display mode (PLAY, PAUSE, etc...) */
      lcd_write_center (3, mode_msg[info->disc_mode], ATTR_NORMAL);
    }

    if (info->disc_mode == CDAUDIO_PLAYING) {
      /*
       * Playing disc info
       */ 
      /* Optimize the display: update modified values */
      if (old_minutes != info->disc_track_time.minutes) {
	sprintf (buf, "%02d", info->disc_track_time.minutes);
	lcd_write (10, 1, buf, ATTR_NORMAL);
      }

      if (old_current_track != info->disc_current_track) {
	sprintf (buf, "%02d", info->disc_current_track);
	lcd_write (16, 1, buf, ATTR_NORMAL);
	
	/* Select new info (text or track length) */
	current_track = info->disc_current_track;
	log_debug (DEBUG_CDPLAYER, "current_track= %d", current_track);
	lcd_init_hscroll ();
	display_track_length ();
      }

      if (old_seconds != info->disc_track_time.seconds) {
	sprintf (buf, "%02d", info->disc_track_time.seconds);
	lcd_write (13, 1, buf, ATTR_NORMAL);
      }

      old_minutes = info->disc_track_time.minutes;
      old_seconds = info->disc_track_time.seconds;
      old_current_track = info->disc_current_track;
    }

    switch (display_type) {

    case DISPLAY_TRACK_TITLE:
      if (current_track > 0) {
	lcd_write_hscroll (-1, 2, disc_data.data_track[current_track-1].track_name);
      }
      break;

    case DISPLAY_CD_ARTIST:
      lcd_write_hscroll (-1, 2, disc_data.data_artist);
      break;

    case DISPLAY_CD_TITLE:
      lcd_write_hscroll (-1, 2, disc_data.data_title);
      break;
    
    default: break;
    }
  }
}

static int check_for_audio_cd ()
{
  register int i;

  first_audio_track = 0;

  for (i = 0 ; i < disc_info.disc_total_tracks ; i++) {
    if (disc_info.disc_track[i].track_type == CDAUDIO_TRACK_AUDIO) {
      first_audio_track = i+1;
      log_debug (DEBUG_CDPLAYER, "first_audio_track= %d", first_audio_track);
      return 0;
    }
  }

  omm_state = OMM_STATE_INIT;
  return -1;
}

/*
 * Main CD check thread
 * check-out CD insert/remove
 */
static void *cd_check_process (void *dummy)
{
  char buf[MAXPATHLEN];

  log_debug (DEBUG_CDPLAYER, "cd_check_process: starting");

  if ((fd_cd = cd_init_device (CD_DEV)) < 0) {
    log_err ("Can't init device %s, exiting.\n", CD_DEV);
    exit (1);
  }

  /* Close tray */
  if (cd_close (fd_cd) < 0) 
    log_err ("Can't close CD");
  else
    tray_closed = TRUE;


  while (1) {
    if (cd_stat (fd_cd, &disc_info) < 0) 
      log_err ("Can't stat CD");

    if (disc_info.disc_present) {
      /* Disc was inserted */
      if (old_disc_present != disc_info.disc_present) {
	log_debug (DEBUG_CDPLAYER, "Disc present!");
	omm_state = STATE_CD;

	/* Check for audio CD */
	if (check_for_audio_cd () < 0) {
	  display_no_audio ();
	}
	else {
	  lcd_write_center (2, "Querying CDDB", ATTR_NORMAL);
	  /* Query CDDB */
	  disc_id = cddb_direct_discid (disc_info);

	  if (cddb_read_disc_data (fd_cd, &disc_data) < 0) {
	    log_debug (DEBUG_CDPLAYER, "Can't access CDDB for %x!", disc_id);

	    /* Default spool directory = disc_id */
	    sprintf (disc_spool_dir, "%s/%s/%s/%x", omm_dir, SPOOL_DIR, EXTRACT_DIR, disc_id);
	  }
	  else {
	    log_debug (DEBUG_CDPLAYER, "CDDB access OK for %x, id= %x", disc_id, disc_data.data_id);
	    if (cddb_write_disc_data (fd_cd, disc_data) < 0)
	      log_err ("Can't write CCDB !");

	    /* New directory = title - artist */
	    if (disc_data.data_title[0] != 0 && disc_data.data_artist[0] != 0)
	      sprintf (disc_spool_dir, "%s/%s/%s/%s - %s", omm_dir, SPOOL_DIR, EXTRACT_DIR, disc_data.data_title, disc_data.data_artist);
	    else
	      sprintf (disc_spool_dir, "%s/%s/%s/%x", omm_dir, SPOOL_DIR, EXTRACT_DIR, disc_id);
	  }

	  /* Create spool dir for current disc */
	  if (mkdir (disc_spool_dir, 0755) < 0)
	    log_err ("mkdir: %s : %s", disc_spool_dir, strerror (errno));
	  else {
	    FILE *fp;

	    /* Create .disc_info */
	    sprintf (buf, "%s/%s", disc_spool_dir, DISC_INFO_FILE);
	    creat (buf, 0644);
	    fp = fopen (buf, "w");
	    if (fp) {
	      fprintf (fp, "%x\n", disc_id);
	      fclose (fp);
	    }
	  }

	  lcd_write_center (2, "Done", ATTR_NORMAL);
	}
      }

      display_disc_info (&disc_info);
    }
    else {
      /* Disc was removed */
      if (old_disc_present != disc_info.disc_present) {
	log_debug (DEBUG_CDPLAYER, "No disc!");
	remove_spool_dir ();
	display_no_disc ();
	omm_state = OMM_STATE_INIT;
      }
    }
	
    usleep (500000);
    cd_check_count++;
    old_disc_present = disc_info.disc_present;
    old_disc_mode = disc_info.disc_mode;

    //    log_debug (DEBUG_CDPLAYER, "state= %d", omm_state);
  }
}

static void cd_player_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  if (ev >= ' ')
    log_debug (DEBUG_CDPLAYER, "cd_player_default_handler: got '%c' in state %02x", ev, omm_state);
  else
    log_debug (DEBUG_CDPLAYER, "cd_player_default_handler: got '%02x' in state %02x", ev, omm_state);
}


/* 
 * Various CD functions 
 */


static void do_stop ()
{
  if (cd_stop (fd_cd) < 0) 
    log_err ("Can't stop CD");
}

static void cd_player_eject_close ()
{
  if (tray_closed) {
    if (cd_eject (fd_cd) < 0) 
      log_err ("Can't eject CD");
    else
      tray_closed = FALSE;
  }
  else {
    if (cd_close (fd_cd) < 0) 
      log_err ("Can't close CD");
    else
      tray_closed = TRUE;
  }
}

static void cd_player_play ()
{
  /* If paused, just resume */
  if (disc_info.disc_mode == CDAUDIO_PAUSED) {
    if (cd_resume (fd_cd) < 0) 
      log_err ("Can't resume CD");
  }
  else {
    if (!disc_info.disc_present) {
      if (cd_close (fd_cd) < 0) 
	log_err ("Can't close CD");
    }

    if (cd_play (fd_cd, current_track) < 0) 
      log_err ("Can't play CD");
  }

  /* Force  time+track redraw */
  old_minutes = old_seconds = old_current_track = -1;
}

static void cd_player_stop ()
{
  do_stop ();

  current_track = first_audio_track;
  display_track_length ();
  lcd_init_hscroll ();
  display_zero_time ();
}

static void cd_player_pause ()
{
  if (cd_pause (fd_cd) < 0) 
    log_err ("Can't pause CD");
}

static void cd_player_next_track ()
{
  int old = current_track;

  while (++current_track <= disc_info.disc_total_tracks) {
    if (disc_info.disc_track[current_track-1].track_type == CDAUDIO_TRACK_AUDIO) {
      select_track ();
      return;
    }
  }
  
  current_track = old;
}

static void cd_player_previous_track ()
{
  int old = current_track;

  while (--current_track >= first_audio_track) {
    if (disc_info.disc_track[current_track-1].track_type == CDAUDIO_TRACK_AUDIO) {
      select_track ();
      return;
    }
  }

  current_track = old;
}

static void cd_player_quit ()
{
  if (disc_info.disc_present && (disc_info.disc_mode == CDAUDIO_PLAYING))
    do_stop ();

  if (cd_finish (fd_cd) < 0) 
    log_err ("Can't finish CD");

  remove_spool_dir ();

  omm_state = OMM_STATE_INIT;
  omm_api_close ();
}

/* Toggle time+track/title display */
static void cd_player_toggle_display ()
{
  /* Next type */
  display_type++;
  if (display_type == DISPLAY_LAST_TYPE)
    display_type = 0;

  switch (display_type) {
  case DISPLAY_CD_LENGTH:
    display_disc_length ();
    break;

  case DISPLAY_TRACK_LENGTH:
    display_track_length ();
    break;

  default:
    lcd_init_hscroll ();
    break;
  }
}

static void display_track_selection ()
{
  char buf[MAXPATHLEN];
  struct stat statb;
  char *track_name = disc_data.data_track[current_track-1].track_name;

  if (track_name && *track_name)
    sprintf (buf, "%s/%d-%s.trk", disc_spool_dir, current_track, track_name);
  else
    sprintf (buf, "%s/%d.trk", disc_spool_dir, current_track);

  if (stat (buf, &statb) < 0)
    lcd_write (4, 0, "   ", ATTR_NORMAL);
  else 
    lcd_write (4, 0, "SEL", ATTR_NORMAL);
}

/* 
 * Mark/Unmark track functions 
 */

static int track_state (int track)
{
  char buf[MAXPATHLEN];
  struct stat statb;
  char *track_name = disc_data.data_track[track-1].track_name;

  if (track_name && *track_name)
    sprintf (buf, "%s/%d-%s.trk", disc_spool_dir, track, track_name);
  else
    sprintf (buf, "%s/%d.trk", disc_spool_dir, track);
 
  if (stat (buf, &statb) < 0)
    return TRACK_UNMARKED;
  else
    return TRACK_MARKED;
}

static int mark_track (int track)
{
  char buf[MAXPATHLEN];
  struct stat statb;
  FILE *fp;
  char *track_name;

  if (track < 1 || track > disc_info.disc_total_tracks)
    return -1;

  track_name = disc_data.data_track[track-1].track_name;

  if (track_name && *track_name)
    sprintf (buf, "%s/%d-%s.trk", disc_spool_dir, track, track_name);
  else
    sprintf (buf, "%s/%d.trk", disc_spool_dir, track);

  if (creat (buf, 0644) > 0)
    return TRACK_MARKED;
  else
    return -1;
}

static int unmark_track (int track)
{
  char buf[MAXPATHLEN];
  char *track_name = disc_data.data_track[track-1].track_name;

  if (track_name && *track_name)
    sprintf (buf, "%s/%d-%s.trk", disc_spool_dir, track, track_name);
  else
    sprintf (buf, "%s/%d.trk", disc_spool_dir, track);

  unlink (buf);

  return TRACK_UNMARKED;
}

static int mark_unmark_track (int track, int action)
{
  if (track < 1 || track > disc_info.disc_total_tracks)
    return -1;

  switch (action) {
  case TRACK_MARK:
    return mark_track (track);

  case TRACK_UNMARK:
    return unmark_track (track);

  case TRACK_TOGGLE:
    if (track_state (track) == TRACK_UNMARKED)
      return mark_track (track);
    else
      return unmark_track (track);

  default: 
    return -1;
  }
}

static void cd_player_mark_unmark_track ()
{
  switch (mark_unmark_track (current_track, TRACK_TOGGLE)) {

  case TRACK_MARKED:
      lcd_write (4, 0, "SEL", ATTR_NORMAL);
      break;

  case TRACK_UNMARKED:
    lcd_write (4, 0, "   ", ATTR_NORMAL);
    break;

  default: break;
  }
}

static void cd_player_mark_unmark_all ()
{
  register int i, action;

  if (track_state (current_track) == TRACK_UNMARKED)
    action = TRACK_MARK;
  else
    action = TRACK_UNMARK;

  cd_player_mark_unmark_track ();

  for (i = 1 ; i <= disc_info.disc_total_tracks ; i++) {
    if (i != current_track)
      mark_unmark_track (i, action);
  }
}

/* Go to a new track */
static void select_track ()
{
  display_track_selection ();
  display_track_length ();
  lcd_init_hscroll ();

  if (disc_info.disc_mode != CDAUDIO_PLAYING) {
    /* If paused, force a stop */
    if (disc_info.disc_mode == CDAUDIO_PAUSED)
      do_stop ();

    /* CD is stopped now */
    display_zero_time ();
  }
  else
    cd_player_play ();
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s [-s]\n", s);
  exit (1);
}

/* Main function */
main (int ac, char **av)
{
  char *cp, *new_omm_dir;

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if ((new_omm_dir = getenv ("OMMDIR")))
    omm_dir = new_omm_dir;

  init_debug ();

  /* Get options */
  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {
      case 's' :
	use_stdin = TRUE; break;

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }

  log_debug (DEBUG_CDPLAYER, "starting.");

  /* Init API */
  omm_api_init (cd_event_handlers, CDPLAYER_FIFO, use_stdin);
  lcd_clear_screen ();
  
  /* Init thread attributes */
  if (pthread_attr_setdetachstate (&thread_attr, PTHREAD_CREATE_DETACHED) != 0)
    log_err ("pthread_attr_setdetachstate: %s", strerror (errno));

  /* Start CD detect thread */
  if (pthread_create (&check_cd_thread, &thread_attr, cd_check_process, NULL) != 0)
    log_err ("pthread_create error for cd_check_process");

  /* Main loop */
  omm_api_main_loop ();

  /* End */
  log_debug (DEBUG_CDPLAYER, "exiting.");
  close_debug ();
}
