#ident "$Id: dpy_curses.c,v 1.2 2000/12/12 10:33:42 pierre Exp $"
/*
 * OMM Display implementation for CURSES library
 */
#include "omm.h"
#include <sys/types.h>
#include <termios.h>
#include <curses.h>

#include "dpy.h"

static int xref, yref;
static pthread_mutex_t write_mutex;
static struct termios old_stdin;
static boolean curses_use_stdin;

/* Init */
void lcd_hw_init (boolean use_stdin)
{
  register int y;

  pthread_mutex_init (&write_mutex, NULL);

  initscr ();
  clear ();
  crmode ();
  noecho ();
  curs_set (0);
  
  xref = (COLS - 22)/2;
  yref = y = 3;

  mvprintw (y++, xref, ".--------------------.");
  mvprintw (y++, xref, "|                    |");
  mvprintw (y++, xref, "|                    |");
  mvprintw (y++, xref, "|                    |");
  mvprintw (y++, xref, "|                    |");
  mvprintw (y++, xref, "`--------------------'");
  refresh ();

  curses_use_stdin = use_stdin;
  
  if (use_stdin)
    raw_mode (fileno(stdin), &old_stdin);
}

/* Write string */
void lcd_hw_write (int x, int y, char *s)
{
  pthread_mutex_lock (&write_mutex);
  mvprintw (y+yref+1, x+xref+1, "%s", s);
  refresh ();
  usleep (20000);
  pthread_mutex_unlock (&write_mutex);
}

/* Close */
void lcd_hw_close ()
{
  tcsetattr(fileno(stdin), TCSANOW, &old_stdin);  
  endwin ();
}

/* Get input */
int lcd_hw_input ()
{
  char c;
  fd_set rmask;
  int ntoread;

  FD_ZERO (&rmask);	
  FD_SET (fd_fifo, &rmask);
  if (curses_use_stdin)
    FD_SET (fileno(stdin), &rmask);

  ntoread = select (fd_fifo+1, &rmask, NULL, NULL, NULL);

  if (ntoread < 0) {
    log_err ("lcd_hw_input: select < 0 (%s) !\n", strerror(errno));
    return -1;
  }
  else if (ntoread == 0) {
    log_err ("lcd_hw_input: select = 0 !\n");
    return -1;
  }
  else if (FD_ISSET (fd_fifo, &rmask)) {
    if (read (fd_fifo, &c, 1) < 0) {
      log_err ("lcd_hw_input: read error on %d", fd_fifo);
      return -1;
    }
    return c;
  }
  else if (FD_ISSET (fileno(stdin), &rmask)) {
    if (read (fileno(stdin), &c, 1) < 0) {
      log_err ("lcd_hw_input: read error on %d", fileno(stdin));
      return -1;
    }

    return c;
  }
}

/* Get win */
unsigned long lcd_hw_get_window ()
{
  return 0L;
}
