#ident "$Id$"

/*
 * List functions
 */
#include "omm.h"
#include "list.h"

/* Member of list */
int omm_list_is_member (omm_list **list, char *s)
{
  register omm_list *plist;

  if (!*list)
    return 0;

  for (plist = *list ; plist ; plist = plist->next) {
    if (!strcmp (plist->str, s))
      return 1;
  }

  return 0;

}

/* Add string to list */
int omm_list_add (omm_list **list, char *s)
{
  omm_list *new, *plist;

  if (omm_list_is_member (list, s))
    return -1;

  if (!(new = (omm_list*)calloc (1, sizeof(omm_list))))
    return -1;

  if (!(new->str = strdup (s)))
    return -1;

  log_debug (DEBUG_GENERAL, "omm_list_del: adding %s", s);

  /* New list */
  if (!*list) {
    *list = new;
    return 0;
  }
    
  /* Go to end of list */
  for (plist = *list ; plist->next ; plist = plist->next)
    ;

  /* Link the cell */
  plist->next = new;
  new->prev = plist;

  return 0;
}

/* Remove string from list */
int omm_list_del (omm_list **list, char *s)
{
  omm_list *prev, *next;
  register omm_list *plist;

  if (!*list)
    return -1;

  if (!omm_list_is_member (list, s))
    return -1;

  log_debug (DEBUG_GENERAL, "omm_list_del: removing %s", s);

  /* Go to end of list */
  for (plist = *list ; plist ; plist = plist->next) {
    if (!strcmp (plist->str, s)) {
      prev = plist->prev;
      next = plist->next;

      if (prev) 
	prev->next = next;
      else
	*list = next;

      if (next)
	next->prev = prev;

      free (plist->str);
      free (plist);

      return 0;
    }
  }

  /* Not found */
  return -1;
}

int omm_list_print (omm_list **list)
{
  omm_list *plist;
  int i = 0;

  if (!list) {
    log_debug (DEBUG_GENERAL, "list is empty!");
    return -1;
  }

  for (plist = *list ; plist ; plist = plist->next)
    log_debug (DEBUG_GENERAL, "list[%d] = %s", i++, plist->str);

  return 0;
}
