#ident "$Id$"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>

#include "misc.h"

/* 
 * Misc functions
 */

/* Set a device in raw mode */
void raw_mode (int fd, struct termios *old_term)
{
    struct termios term;

    tcgetattr(fd, &term);
    
    tcgetattr(fd, old_term);
    term.c_iflag = IGNBRK;
    term.c_lflag = 0;
    term.c_oflag = 0;
    //    term.c_cflag |= (CREAD | CRTSCTS);
    term.c_cc[VMIN] = 1;
    term.c_cc[VTIME] = 0;
    term.c_iflag &= ~(IXON|IXOFF|IXANY);
    term.c_cflag &= ~(PARENB | CSIZE);
    term.c_cflag |= CS8;
    term.c_cflag &= ~CLOCAL;

    tcsetattr(fd, TCSANOW, &term);
}

/* Remove suffix from filename */
void remove_suffix (char *name, const char *suffix)
{
  char *np;
  const char *sp;

  np = name + strlen (name);
  sp = suffix + strlen (suffix);

  while (np > name && sp > suffix)
    if (*--np != *--sp)
      return;
  if (np > name) {
    if (*(np-1) == '.')
      *(np-1) = 0;
    else
      *np = '\0';
  }
}

/* Check if filename contains suffix */
int check_suffix (char *name, char *suffix)
{
  char *np;

  if ((np = strchr (name, '.'))) {
    if (!strcmp (np+1, suffix))
      return 1;
  }
  
  return 0;
}

/* Open a FIFO and return fd */
int open_fifo (char *fifo_name)
{
  int fd;

  /* Create if not existing */
  if (mknod (fifo_name, S_IFIFO | 0666, 0) != 0  && errno != EEXIST) {
    log_err ("%s: %s", fifo_name, strerror(errno));
    exit (1);
  }

  if ((fd = open (fifo_name, O_RDWR)) < 0) {
    log_err ("%s: %s", fifo_name, strerror(errno));
    exit (1);
  }

  return fd;
}

/* Remove file with error test */
int my_unlink (const char *name)
{
  if (unlink (name) < 0) {
    log_err ("my_unlink: %s: %s", name, strerror(errno));
    return -1;
  }
  else 
    return 0;
}

/* Remove dir with error test */
int my_rmdir (const char *name)
{
  if (rmdir (name) < 0) {
    log_err ("my_rmdir: %s: %s", name, strerror(errno));
    return -1;
  }
  else 
    return 0;
}
