#ident "$Id: omm_api.c,v 1.2 2000/12/12 10:38:34 pierre Exp pierre $"
/*
 * OMM API
 */

#include "omm.h"

static char omm_api_end;
static omm_event_handler *p_ev_tab;
static omm_api_event current_event;
static void (*fn_input_ok)();
static void (*fn_input_abort)();

static void lazy () 
{}

/* Public functions */

/* Init the API */
void omm_api_init (omm_event_handler *event_handler_tab, char *fifo_name, boolean use_stdin)
{
  char buf[MAXPATHLEN];

  fn_input_ok = lazy;
  fn_input_abort = lazy;
  p_ev_tab = event_handler_tab;
  omm_state = OMM_STATE_INIT;
  
  fd_fifo = open_fifo (fifo_name);
  sprintf (buf, "%s_reply", fifo_name);
  fd_fifo_reply = open_fifo (buf);

  lcd_init (use_stdin);
}

/* Close the API */
void omm_api_close (void)
{
  omm_api_end = 1;
  lcd_close ();
  close (fd_fifo);
}

char *omm_api_get_input_str ()
{
  return omm_input_str;
}

/* Start the input char process */
void omm_api_start_input (char *s, void (*fn_ok), void (*fn_abort))
{
  omm_input_ptr = 0;
  memset ((void*)omm_input_str, 0, sizeof(omm_input_str));

  fn_input_ok = fn_ok;
  fn_input_abort = fn_abort;
  
  lcd_write_center (1, s, ATTR_NORMAL);
  lcd_clear (0, 2, LCD_WIDTH-1, 3);
  lcd_write (0, 3, "^", ATTR_NORMAL);
  omm_state = OMM_STATE_INPUT;
}

void omm_api_main_loop (void)
{
  omm_api_event ev = 0;

  while (!omm_api_end) {
    omm_api_get_event (&ev);
    omm_api_handle_event (&ev);
  }
}

void omm_api_get_event (omm_api_event *ev)
{
  current_event = *ev = lcd_input ();
}

void omm_api_handle_event (omm_api_event *ev)
{
  register omm_event_handler *p = p_ev_tab;

  switch (omm_state) {

  case OMM_STATE_INPUT:
    switch (*ev) {
    case '\n':
      lcd_write (omm_input_ptr, 3, " ", ATTR_NORMAL);
      omm_input_str[++omm_input_ptr] = 0;
      omm_state = OMM_STATE_INIT;
      fn_input_ok ();
      log_debug (DEBUG_GENERAL, "input str validated!");
      break;

    case 0x06: /* Right */
      if (!omm_input_str[omm_input_ptr])
	omm_input_str[omm_input_ptr] = ' ';

      lcd_write (omm_input_ptr, 3, " ", ATTR_NORMAL);
      lcd_write (++omm_input_ptr, 3, "^", ATTR_NORMAL);
      break;

    case 0x02: /* Left */
      if (omm_input_ptr) {
	omm_input_str[omm_input_ptr] = 0;
	lcd_clear_line (2);
	lcd_write (0, 2, omm_input_str, ATTR_NORMAL);
	lcd_write (omm_input_ptr, 3, " ", ATTR_NORMAL);
	lcd_write (--omm_input_ptr, 3, "^", ATTR_NORMAL);
      }
      break;

    case 0x1b: /* ESC: abort input */
      omm_input_ptr = 0;
      memset ((void*)omm_input_str, 0, sizeof(omm_input_str));
      omm_state = OMM_STATE_INIT;
      fn_input_abort ();

      break;

    default:
      omm_input_str[omm_input_ptr] = *ev;
      lcd_write (0, 2, omm_input_str, ATTR_NORMAL);
      break;
    }

    log_debug (DEBUG_ENCODER, "current input str= '%s'", omm_input_str);
    break;

  default:
    while (p->code) {
      if (p->code == *ev && (omm_state == p->current_state)) {
	int old_state = omm_state;

	/* Call handler and return */
	(*p->handler) ();
	return;
      }
    
      p++;
    }

    /* Execute handler event if any */
    if (p->handler && ((omm_state == p->current_state) || p->current_state == OMM_ANY_STATE))
      (*p->handler) ();

  }
}

omm_api_event omm_api_get_current_event ()
{
  return current_event;
}
