#ident "$Id: encoder.c,v 1.4 2000/12/20 07:47:39 pierre Exp pierre $"
/*
 * OMM Encoder
 */

#define EXTERN
#include "omm.h"
#include "encoder.h"

static char *omm_dir = DEFAULT_OMMDIR;
static char *new_cdparanoia, *cdparanoia = CDPARANOIA;
static char *new_mp3enc, *mp3enc = MP3ENC;

static struct disc_info disc_info;
static int pipe_in[2];
static int paranoia_pid;
static int mp3e_pid;
static pthread_t display_status_thread;
static pthread_t explore_dir_thread;
static pthread_t trackname_thread;
static char action_end;
static char disc_name[1024];
static char full_track_name[1024];
static char current_file_name[1024];
static int comp_file_name_ptr;
static int current_disc_id, disc_id;
static int fd_cd;
static char disc_not_here;
static char *scrolled_str;
static int action_error;
static int track_count = 0;
static char *comp_file_name;
static pthread_mutex_t scroll_mutex;

/* States handling definition */
static omm_event_handler encoder_event_handlers[] = {
  'c', STATE_INSERT_CD, encoder_close_cd,	/* Close the tray */
  'q', OMM_STATE_INIT, encoder_quit,		/* Quit */
  'q', STATE_INSERT_CD, encoder_quit,		/* Quit */
  'p', OMM_STATE_INIT, encoder_start,		/* Start */
  's', OMM_STATE_INIT, encoder_stop,		/* Stop */
  's', STATE_INSERT_CD, encoder_stop,		/* Stop */
  'B', OMM_STATE_INIT, lcd_bg_screen,
  'B', STATE_INSERT_CD, lcd_bg_screen,
  'F', OMM_STATE_INIT, lcd_fg_screen,
  'F', STATE_INSERT_CD, lcd_fg_screen,
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, encoder_default_handler,
  0, 0, NULL
};

static void kill_programs ()
{
  if (paranoia_pid) {
    kill (paranoia_pid, SIGTERM);
    paranoia_pid = 0;
  }

  if (mp3e_pid) {
    kill (mp3e_pid, SIGTERM);
    mp3e_pid = 0;
  }
}

static void clear_screen ()
{
  lcd_clear_screen ();
  lcd_write (0, 0, "ENC", ATTR_NORMAL);
}

static void set_comp_file_name ()
{
  comp_file_name = omm_api_get_input_str ();
}

static void set_comp_file_name_start ()
{
  set_comp_file_name ();
  encoder_start ();
}

static void abort_comp_file_name_input ()
{
  lcd_clear_screen ();
  lcd_write (0, 0, "ENC", ATTR_NORMAL);
  lcd_write_center (2, "Input aborted!", ATTR_NORMAL);
}

/* Quit encoder */
static void encoder_quit ()
{
  kill_programs ();
  omm_state = OMM_STATE_INIT;
  omm_api_close ();
}


static void got_sigchld (int sig)
{
    int r;

    log_debug (DEBUG_ENCODER, "Got SIGCHLD!");

    while (wait (&r) > 0)
      ;
}

static void got_interrupt (int sig)
{	
  log_debug (DEBUG_ENCODER, "Got interrupt %d!", sig);

  encoder_quit ();
  exit (1);
}


static void read_program_status (char *s, int size)
{
  char c = 0;
  int cnt = 0;

  //  log_debug (DEBUG_ENCODER, "read_program_status: %d bytes to read", size);

  memset ((void*)s, 0, size);
  while (c != '\n' && cnt < size) {
    if (read (PROGRAM_IN, &c, 1) < 0) {
      log_err ("read_program_status: %s: %s", s, strerror(errno));
      action_end = 1;
    }
    
    if (c != '\n') {
      *(s+cnt) = c;
      cnt++;
    }
  }

  //  log_debug (DEBUG_ENCODER, "read_program_status: %d bytes read", cnt);
}

/* Check extract dir to see if not empty (nothing to do...) */
static int check_extract_dir_size ()
{
  char buf[MAXPATHLEN];
  DIR *dirfile;
  struct dirent *anentry;
  int count = 0;

  sprintf (buf, "%s/%s/%s", omm_dir, SPOOL_DIR, EXTRACT_DIR);
  dirfile = opendir(buf);

  while ((anentry = readdir (dirfile)) != NULL) {
    /* skip some files */
    if (anentry->d_name[0] != '.') 
      count++;
  }

  closedir (dirfile);
  return count;
}

static void *display_scrolled_process (void *dummy)
{
  for (;;) {
    pthread_mutex_lock (&scroll_mutex);
    if (scrolled_str)
      lcd_write_hscroll (-1, 1, scrolled_str);
    pthread_mutex_unlock (&scroll_mutex);
    usleep (500000);
  }
}


/* Display info from forked program */
static void *display_status_process (void *dummy)
{
  char statusbuf[256], buf[LCD_WIDTH+1];
  int errcode = 0;

  log_debug (DEBUG_ENCODER, "display_status_process: starting");

  while (!action_end) {
    read_program_status (statusbuf, sizeof(statusbuf));

    if (statusbuf[1] == 'X')
      break;
    else if (statusbuf[1] == 'E') {
      sscanf (&statusbuf[3], "%x", &errcode);
      action_end = 1;
      break;
    }
    else {
      lcd_write_center (3, &statusbuf[3], ATTR_NORMAL);
    }
  }

  log_debug (DEBUG_ENCODER, "display_status_process: exiting, action_end= %d, errcode= %d", action_end, errcode);

  if (action_end) {
    if (errcode != 0) {
      sprintf (buf, "ERROR %d", errcode);
      lcd_write_center (2, buf, ATTR_NORMAL);
    }
    else
      lcd_write_center (2, "ABORTED", ATTR_NORMAL);

    kill_programs ();
    pthread_exit ((void*)1);
  }
  else {
    lcd_write_center (2, "DONE", ATTR_NORMAL);
    pthread_exit ((void*)0);
  }
}

/* 
 * Extract a track. First param is the track number, 2nd is the track name
 */
static void extract_track (char *track_name)
{
  char buf[LCD_WIDTH+1], track[3];

  memset ((void*)track, 0, sizeof(track));
  memset ((void*)full_track_name, 0, sizeof(full_track_name));
  lcd_init_hscroll ();

  sprintf (full_track_name, "%s - %s", track_name, disc_name);
  sprintf (current_file_name, "%s.wav", track_name);

  /* Get track # */
  track[0] = *track_name;
  if (isdigit (*(track_name+1)))
    track[1] = *(track_name+1);

  log_debug (DEBUG_ENCODER, "extract_track: track= %s", track);

  pthread_mutex_lock (&scroll_mutex);
  scrolled_str = full_track_name;
  pthread_mutex_unlock (&scroll_mutex);

  log_debug (DEBUG_ENCODER, "extract_track: full_track_name= %s", full_track_name);

  lcd_write_center (2, "EXTRACTING", ATTR_NORMAL);

  /* Start status thread */
  if (pthread_create (&display_status_thread, NULL/*&thread_attr*/, display_status_process, NULL) != 0)
    log_err ("pthread_create error for display_status_process");

  /* Start extractor */
  if (!(paranoia_pid = fork ())) {
    log_debug (DEBUG_ENCODER, "extract_track: forked %s pid = %d", cdparanoia, getpid());

    dup2 (pipe_in[1], 1);
    close (pipe_in[1]);

    if (execlp (cdparanoia, cdparanoia, track, current_file_name, NULL) < 0) {
      log_err ("execlp: %s: %s", cdparanoia, strerror(errno));
      encoder_quit ();
    }
  }
}

/* 
 * Encode an audio file generated by CDPARANOIA. 
 */
static void encode_file (char *file_name)
{
  char buf[LCD_WIDTH+1];

  memset ((void*)full_track_name, 0, sizeof(full_track_name));
  lcd_init_hscroll ();

  /* Set the track name + wav file name */
  sprintf (full_track_name, "%s - %s", file_name, disc_name);
  sprintf (current_file_name, "%s.wav", file_name);

  pthread_mutex_lock (&scroll_mutex);
  scrolled_str = full_track_name;
  pthread_mutex_unlock (&scroll_mutex);

  log_debug (DEBUG_ENCODER, "encode_track: full_track_name= %s", full_track_name);
  log_debug (DEBUG_ENCODER, "encode_track: wav file= %s", current_file_name);

  lcd_write_center (2, "ENCODING", ATTR_NORMAL);

  /* Start status thread */
  if (pthread_create (&display_status_thread, NULL /*&thread_attr*/, display_status_process, NULL) != 0)
    log_err ("pthread_create error for display_status_process");

  /* Start encoder */
  if (!(mp3e_pid = fork ())) {
    log_debug (DEBUG_ENCODER, "encode_track: forked %s pid = %d", mp3enc, getpid());

    dup2 (pipe_in[1], 1);
    close (pipe_in[1]);

    if (execlp (mp3enc, mp3enc, current_file_name, NULL) < 0) {
      log_err ("execlp: %s: %s", mp3enc, strerror(errno));
      encoder_quit ();
    }
  }
}

static void encoder_close_cd ()
{
  if (cd_close (fd_cd) < 0) 
    log_err ("Can't close CD");
}

static void ask_for_disc ()
{
  int status;

  lcd_clear_screen ();

  pthread_mutex_lock (&scroll_mutex);
  scrolled_str = disc_name;
  pthread_mutex_unlock (&scroll_mutex);

  lcd_write (0, 0, "ENC", ATTR_NORMAL);
  lcd_write_center (2, "INSERT DISC!", ATTR_NORMAL);
  cd_eject (fd_cd);
  omm_state = STATE_INSERT_CD;
}

static void explore_directory (char *dir, int action)
{
  DIR *dirfile;
  struct dirent *anentry;
  struct stat statbuf;
  char buf[MAXPATHLEN];
  FILE *fp;
  void *join_return;

  log_debug (DEBUG_ENCODER, "explore_directory: %s action= %d", dir, action);

  chdir (dir);
  dirfile = opendir(".");

  while ((anentry = readdir (dirfile)) != NULL && !action_end) {
    /* skip some files */
    if (anentry->d_name[0] == '.' /*|| check_suffix (anentry->d_name, MP3_SUFFIX)*/ || (action == ACTION_EXTRACT && check_suffix (anentry->d_name, WAV_SUFFIX)) || stat(anentry->d_name, &statbuf) < 0) {
      /* 
       * If EXTRACT, get disc Id from .disc_info file and compare it
       * with the current disc (if any). If does not match, ask for the right
       * disc.
       */ 
      if (!strcmp (anentry->d_name, DISC_INFO_FILE)) {
	if (action == ACTION_EXTRACT) {
	  if ((fp = fopen (anentry->d_name, "r"))) {
	    fscanf (fp, "%x", &current_disc_id);
	    fclose (fp);
	    log_debug (DEBUG_ENCODER, "current_disc_id= %x", current_disc_id);

	    /* Check for matching disc_id */
	    if ((fd_cd = cd_init_device (CD_DEV)) < 0) {
	      log_err ("Can't init device %s, exiting.\n", CD_DEV);
	      encoder_quit ();
	    }

	    lcd_init_hscroll ();

	    /* Wait for correct disc */
	    for (;;) {
	      if (cd_stat (fd_cd, &disc_info) < 0) {
		log_err ("Can't stat CD");
		encoder_quit ();
	      }

	      /* Disc is here, check for Id */
	      if (disc_info.disc_present) {
		disc_not_here = 0;
	      
		disc_id = cddb_direct_discid (disc_info);

		/* Check for matching disc_id */
		if (disc_id == current_disc_id) {
		  if (cd_finish (fd_cd) < 0) 
		    log_err ("Can't finish CD");
		  lcd_write (0, 0, "ENC", ATTR_NORMAL);
		  omm_state = OMM_STATE_INIT;
		  break;
		}
		else {
		  /* Not the right disc */
		  log_debug (DEBUG_ENCODER, "Not the right disc, ask for it!");
		  ask_for_disc ();
		}
	      }
	      else {
		/* Disc is not present */
		if (!disc_not_here) {
		  log_debug (DEBUG_ENCODER, "Disc not here, ask for it!");
		  ask_for_disc ();
		  disc_not_here = 1;
		}
	      }

	      sleep (1);
	    }
	  }
	  else {
	    log_err ("Can't read %s, exiting.", DISC_INFO_FILE);
	    encoder_quit ();
	  }
	}
	/* ACTION_ENCODE */
	else {
	  /* Check error to remove .disc_info */
	  if (action_error == 0)
	    my_unlink (anentry->d_name);
    	}
      }
      else {
	//	log_debug (DEBUG_ENCODER, "skipped %s !", anentry->d_name);
      }

      /* Next file */
      continue;
    }
    /*
     * If directory (ie disc name), recursively explore it...
     */
    else if (S_ISDIR (statbuf.st_mode)) {
      memset ((void*)disc_name, 0, sizeof(disc_name));
      strcpy (disc_name, anentry->d_name);
      log_debug (DEBUG_ENCODER, "=== disc_name= %s", disc_name);
      explore_directory (anentry->d_name, action);

      /* If OK, remove directory */
      if (action == ACTION_ENCODE && action_error == 0) {
	log_debug (DEBUG_ENCODER, "Removing dir \"%s\"", disc_name);
	my_rmdir (disc_name);
      }
    }
    /*
     * Plain file, should be a .trk or a .wav
     */
    else {
      /* .trk file */
      if (!check_suffix (anentry->d_name, WAV_SUFFIX)) {

	log_debug (DEBUG_ENCODER, "track '%s' to be EXTRACTED", anentry->d_name);
	/* Remove .trk suffix */
	remove_suffix (anentry->d_name, TRACK_SUFFIX);
      }
      /* File is a .wav (action should be ACTION_ENCODE) */
      else {
	log_debug (DEBUG_ENCODER, "file '%s' to be ENCODED", anentry->d_name);

	/* Remove .wav suffix */
	remove_suffix (anentry->d_name, WAV_SUFFIX);
      }

      /* EXTRACT or ENCODE the track */
      if (action == ACTION_EXTRACT)
	extract_track (anentry->d_name);
      else
	encode_file (anentry->d_name);

      /* Wait for action to be done */
      log_debug (DEBUG_ENCODER, "waiting track %s", anentry->d_name); 
      pthread_join (display_status_thread, &join_return);
      log_debug (DEBUG_ENCODER, "track '%s' is done %d", anentry->d_name, (int)join_return); 

      if (action == ACTION_EXTRACT) {
	if ((int)join_return == 0) {
	  /*
	   * If extracting OK, we can remove the track # and just keep the
	   * .wav
	   */
	  sprintf (buf, "%s.%s", anentry->d_name, TRACK_SUFFIX);
	  log_debug (DEBUG_ENCODER, "Extract OK, removing '%s' !", buf);
	  my_unlink (buf);
	}
	else {
	  /*
	   * If extracting aborted, we should remove the bad .wav and keep the
	   * track #
	   */   
	  log_debug (DEBUG_ENCODER, "Extract ABORTED, '%s.%s' NOT removed, '%s' removed", anentry->d_name, TRACK_SUFFIX, current_file_name);
	  my_unlink (current_file_name);
	  action_error++;
	}
      }
      /* Encoding */
      else {
	if ((int)join_return == 0) {
	  /*
	   * If encoding OK, we can remove the .wav and move the .mp3 to the
	   * compilation dir
	   */   
	  log_debug (DEBUG_ENCODER, "removing '%s'", current_file_name); 
	  my_unlink (current_file_name);

	  remove_suffix (current_file_name, WAV_SUFFIX);
	  sprintf (buf, "mv \"%s.%s\" \"%s/%s/%s/%s - %s.%s\"", current_file_name, MP3_SUFFIX, omm_dir, MP3_DIR, comp_file_name, current_file_name, disc_name, MP3_SUFFIX);
	  system (buf);
	}
	else {
	  action_error++;
	  
	  /* Encoding aborted ==> remove the .mp3 and keep the .wav */
	  remove_suffix (current_file_name, WAV_SUFFIX);
	  sprintf (buf, "%s.%s", current_file_name, MP3_SUFFIX);
	  log_debug (DEBUG_ENCODER, "removing '%s'", buf); 
	  my_unlink (buf);
	}
      }
    }
  }

  chdir ("..");
}

static void *explore_dir_process (void *dummy)
{
  char buf[MAXPATHLEN];

  action_error = 0;

  /* Create target directory in MP3 hierarchy */
  sprintf (buf, "%s/%s/%s", omm_dir, MP3_DIR, comp_file_name);
  mkdir (buf, 0755);

  /* Extract + encode tracks */
  sprintf (buf, "%s/%s/%s", omm_dir, SPOOL_DIR, EXTRACT_DIR);

  /* Extract */
  explore_directory (buf, ACTION_EXTRACT);

  /* Encode */
  explore_directory (buf, ACTION_ENCODE);

  pthread_exit (0);
}

/* Start encoding */
static void encoder_start ()
{
  action_end = 0;

  if (!comp_file_name || *comp_file_name == 0) {
    omm_api_start_input ("Compilation name?", set_comp_file_name_start, abort_comp_file_name_input);

    return;
  }

  if (track_count) {
    if (pthread_create (&explore_dir_thread, &thread_attr, explore_dir_process, NULL) != 0)
      log_err ("pthread_create error for explore_dir_process");
  }
}

/* Stop encoding */
static void encoder_stop ()
{
  action_end = 1;
}

/* Default input handler */
static void encoder_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  if (ev >= ' ')
    log_debug (DEBUG_ENCODER, "encoder_default_handler: got '%c' in state %02x", ev, omm_state);
  else
    log_debug (DEBUG_ENCODER, "encoder_default_handler: got '%02x' in state %02x", ev&255, omm_state);
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s\n", s);
  exit (1);
}

/* Main function */
main (int ac, char **av)
{
  char *cp, *new_omm_dir;

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if ((new_omm_dir = getenv ("OMMDIR")))
    omm_dir = new_omm_dir;

  if ((new_cdparanoia = getenv ("CDPARANOIA")))
    cdparanoia = new_cdparanoia;

  if ((new_mp3enc = getenv ("MP3ENC")))
    mp3enc = new_mp3enc;
  

  init_debug ();

  /* Get options */
  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }

  log_debug (DEBUG_ENCODER, "starting.");

  signal (SIGINT, got_interrupt);
  signal (SIGQUIT, got_interrupt);
  signal (SIGTERM, got_interrupt);
  signal (SIGCHLD, got_sigchld);

  if (pipe (pipe_in) < 0) {
    log_err ("pipe_in: %s", strerror(errno));
    exit (1);
  }

  /* Init API */
  omm_api_init (encoder_event_handlers, ENCODER_FIFO, FALSE);

  lcd_clear_screen ();
  lcd_write (0, 0, "ENC", ATTR_NORMAL);

  if ((track_count = check_extract_dir_size ()) == 0) {
    lcd_write_center (1, "No selected track!", ATTR_NORMAL);
  }
  else
    omm_api_start_input ("Compilation name?", set_comp_file_name, abort_comp_file_name_input);

  
  /* Create str display process */
  pthread_mutex_init (&scroll_mutex, NULL);
  if (pthread_create (&trackname_thread, &thread_attr, display_scrolled_process, NULL) != 0)
    log_err ("pthread_create error for display_scrolled_process");

  /* Main loop */
  omm_api_main_loop ();

  /* End */
  log_debug (DEBUG_ENCODER, "exiting.");
  close_debug ();
}

