#ident "$Id: filesel.c,v 1.2 2000/12/12 10:28:51 pierre Exp pierre $"
/*
 * filesel.c - the file selector.
 * Inspired by the filesel of zgsmplay
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <curses.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "omm.h"
#include "fileselector.h"

#define MAX_FLEN	LCD_WIDTH-2
#define BARWIDTH	(MAX_FLEN+2)
#define XSIZ		((COLS)/BARWIDTH)
#define YSIZ		LCD_HEIGHT

#define fwinxpos(f)	(BARWIDTH*((f-1)/YSIZ))
#define fwinypos(f)	(((f)-1)%YSIZ)

struct mp3dir_tag {
  char name[256];
  char isdir;            /* 0=file, 1=dir. */
  char marked;
  };

int mp3dirsiz;
struct mp3dir_tag *mp3dir=NULL;
int mp3dir_byte_size=64*sizeof(struct mp3dir_tag);
int mp3dir_byte_incr=32*sizeof(struct mp3dir_tag);

int tagview_mode=0;
int nodelprompt=0;

static char my_filename[MAXPATHLEN];
static char cdir[MAXPATHLEN+1];
static int curent;
static int oldent, startfrom, oldstart, f, markchange;
static omm_list *my_list;
static int ysel;
static char *current_filename;
static pthread_t filename_thread;
static pthread_mutex_t filename_mutex;
static char *mp3dir_base;
static char *mp3_suffix;
static FILE *fp_out;
static char *the_outfile;

static void *display_filename_process (void *dummy)
{
  for (;;) {
    pthread_mutex_lock (&filename_mutex);
    if (current_filename)
      lcd_write_hscroll (2, ysel, current_filename);
    pthread_mutex_unlock (&filename_mutex);
    usleep (500000);
  }
}

/* allocate some initial space for the mp3dir[] array.
 * This should be run before starting up curses.
 * (It's called by main(), in fact.)
 */
void mp3dir_init()
{
  if(mp3dir!=NULL) return;	/* sanity check */

  if((mp3dir=malloc(mp3dir_byte_size))==NULL) {
    log_err ("filesel: out of memory!");
    exit(1);
  }
}


/* make mp3dir bigger if needed.
 * call this *before writing each new entry*.
 * if this returns zero, the caller should quit with not-enough-memory
 * error, exiting curses with screenoff() if needed.
 *
 * note that mp3dir is never made smaller. This would complicate things
 * considerably for a fairly small gain, I think.
 */
int mp3dir_resize_if_needed(int newent)
{
  /* this is absurdly conservative, just to be on the safe side :-) */
  if((newent+1)*sizeof(struct mp3dir_tag)>=mp3dir_byte_size)
    {
      mp3dir_byte_size+=mp3dir_byte_incr;
      if((mp3dir=realloc(mp3dir,mp3dir_byte_size))==NULL)
	return(0);		/* out of memory */
    }

  return(1);		/* worked */
}


void printmax(char *str,int maxlen)
{
  static char buf[256];

  strcpy(buf,str);
  buf[maxlen]=0;
  lcd_addstr(buf, ATTR_NORMAL);
}


void printmaxspc(char *str,int maxlen)
{
  static char buf[256];
  int f;

  for(f=0;f<strlen(str);f++)
    buf[f]=' ';
  buf[f]=0;

  buf[maxlen]=0;
  lcd_addstr(buf, ATTR_NORMAL);
}


/* this could also be file not found of course */
int permissiondenied(char *fname)
{
  FILE *junk;
  if((junk=fopen(fname,"rb"))==NULL)
    return(1);
  fclose(junk);
  return(0);
}


void prettyfile(char *buf,struct mp3dir_tag *mp3dptr)
{
  if(mp3dptr->isdir)
    {
      buf[0]='(';
      strcpy(buf+1,mp3dptr->name);
      strcat(buf,")");
    }
  else
    strcpy(buf,mp3dptr->name);
}


void showbar(int entnum,int selected,int startfrom)
{
  char ctmp[100];
  int xpos,ypos;

  xpos=fwinxpos(entnum-startfrom+1);
  ypos=fwinypos(entnum-startfrom+1);
  prettyfile(ctmp,&(mp3dir[entnum]));
  lcd_move(xpos,ypos);

   if(mp3dir[entnum].isdir) 
    lcd_addch('D', ATTR_NORMAL); 
   else if(mp3dir[entnum].marked) 
    lcd_addch('*', ATTR_NORMAL); 
  else 
    lcd_addch(' ', ATTR_NORMAL);

   if(selected) {
     ysel = ypos;
     lcd_init_hscroll ();
     current_filename = (char*)mp3dir[entnum].name;

     log_debug (DEBUG_FILESEL, "ysel= %d, file= %s", ysel, current_filename);
    lcd_addch('>', ATTR_NORMAL); 
   }
  else 
    lcd_addch(' ', ATTR_NORMAL);
  printmax(mp3dir[entnum].name,MAX_FLEN);
}



int load_single_file(int curent)
{
  fprintf (fp_out, "%s/%s.%s\n", cdir, mp3dir[curent].name, mp3_suffix);
}


void showmp3dir(int startfrom,int unshow)
{

  static char ctmp[1024];
  int f,ypos,xpos;

  getcwd(cdir,MAXPATHLEN);

  for(f=startfrom;f<=mp3dirsiz;f++)
    {
      xpos=fwinxpos(f-startfrom+1);
      if(xpos+BARWIDTH>LCD_WIDTH) break;
      ypos=fwinypos(f-startfrom+1);
      prettyfile(ctmp,&(mp3dir[f]));

      lcd_move(xpos,ypos);
      
      if(mp3dir[f].marked && !unshow) 
	lcd_addch('*', ATTR_NORMAL); 
      else {
	if (unshow || !mp3dir[f].isdir)
	  lcd_addch(' ', ATTR_NORMAL);
	else
	  lcd_addch('D', ATTR_NORMAL);
	
      }

      lcd_addch(' ', ATTR_NORMAL);

      if(unshow)
	printmaxspc(mp3dir[f].name,MAX_FLEN);
      else
	printmax(mp3dir[f].name,MAX_FLEN);
    }
}


int gcompare(void *gn1,void *gn2)
{
  struct mp3dir_tag *g1,*g2;
  char buf1[256],buf2[256];

  g1=(struct mp3dir_tag *)gn1; g2=(struct mp3dir_tag *)gn2;
  prettyfile(buf1,g1); prettyfile(buf2,g2);

  return(strcmp(buf1,buf2));
}


void readmp3dir()
{
  DIR *dirfile;
  struct dirent *anentry;
  struct stat buf;
  int entnum,l,isdir;
  char suffix[MAXPATHLEN];

  getcwd(cdir,MAXPATHLEN);
  dirfile=opendir(".");
  entnum=0;
  while((anentry=readdir(dirfile))!=NULL)
    {
      if(anentry->d_name[0]=='.' /*&& anentry->d_name[1]!='.'*/ )
	continue;	/* skip (most) 'dot' files */
      if((strcmp(anentry->d_name,".")!=0) &&	/* no . and no .. if at root */
	 (!((strcmp(cdir,"/")==0)&&(strcmp(anentry->d_name,"..")==0))))
	{
	  if((stat(anentry->d_name,&buf))==-1)
	    buf.st_mode=0;
	  isdir=S_ISDIR(buf.st_mode);
	  /* directories, mp3 tested here */
	  if(((l=strlen(anentry->d_name))>4) || isdir)
	    /* Should be configurable */
	    if((!strcasecmp(anentry->d_name+l-3, mp3_suffix)) || isdir)
	      {
		entnum++;
		if(!mp3dir_resize_if_needed(entnum))
		  {
		    //          screenoff();
		    log_err ("filesel: out of memory");

		    exit(1);
		  }
        
		//		sprintf (suffix, ".%s", mp3_suffix);
		remove_suffix (anentry->d_name, mp3_suffix);
		strcpy(mp3dir[entnum].name,anentry->d_name);
		mp3dir[entnum].isdir=isdir;

		sprintf (my_filename, "%s/%s", cdir, mp3dir[entnum].name);
		if (omm_list_is_member (&my_list, my_filename))
		  mp3dir[entnum].marked=1;
		else
		  mp3dir[entnum].marked=0;

	      }
	}
    }

  closedir(dirfile);
  mp3dirsiz=entnum;

  log_debug (DEBUG_FILESEL, "dirsiz= %d", mp3dirsiz);

  qsort(&(mp3dir[1]),mp3dirsiz,sizeof(struct mp3dir_tag),(void *)gcompare);
}


void redrawall(int curent,int startfrom)
{
  showmp3dir(startfrom,0);
  showbar(curent,1,startfrom);
}


int delete_file(char *filename)
{
  char buf[MAXPATHLEN];
  /* XXX NYI */
  sprintf (buf, "%s.%s", filename, mp3_suffix);
  log_debug (DEBUG_FILESEL, "delete_file: %s", buf);
  if (my_unlink (buf) == 0)
    return(1);
  else
    log_err ("delete_file: %s: %s", buf, strerror(errno));
}



void filesel_before ()
{
  oldent = curent;
  markchange = 0;
}

void filesel_update ()
{
  oldstart=startfrom;
  while(curent<startfrom)
    startfrom-=YSIZ;
  while(fwinxpos(curent-startfrom+1)+BARWIDTH>LCD_WIDTH)
    startfrom+=YSIZ;

  if(startfrom<1) startfrom=1;

  if(startfrom!=oldstart)
    {
    showbar(oldent,0,oldstart);
    showmp3dir(oldstart,1);
    showmp3dir(startfrom,0);
    showbar(curent,1,startfrom);
    }
  else  
    if(oldent!=-1 && (curent!=oldent || markchange))
      {
      showbar(oldent,0,startfrom);
      showbar(curent,1,startfrom);
      }

}

void filesel_init (char *suffix, char *base_dir, char *outfile)
{
  if (outfile) {
    the_outfile = outfile;
    fp_out = fopen (outfile, "w");
    if (!fp_out) {
      log_err ("filesel_init: Can't open %s: %s", outfile, strerror(errno));
      fp_out = stderr;
    }
  }
  else
    fp_out = stderr;

  if (base_dir)
    mp3dir_base = strdup (base_dir);
  else
    mp3dir_base= getenv ("MP3DIR");

  if (suffix)
    mp3_suffix = strdup (suffix);
  else
    mp3_suffix = "mp3";

  if (mp3dir_base)
    chdir (mp3dir_base);

  mp3dir_init ();
  curent = 1;
  oldent = 1;
  startfrom = 1;

  readmp3dir();
  showmp3dir(startfrom, 0);
  showbar(curent, 1, startfrom);

  pthread_mutex_init (&filename_mutex, NULL);

  if (pthread_create (&filename_thread, &thread_attr, display_filename_process, NULL) != 0)
    log_err ("pthread_create error for display_filename_process");
}

void filesel_quit ()
{
  omm_state = OMM_STATE_INIT;
  omm_api_close ();
  if (fp_out != stderr)
    fclose (fp_out);
}


void filesel_redraw ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  readmp3dir();
  if(curent>mp3dirsiz) curent=mp3dirsiz;
  oldent=curent;
  redrawall(curent,startfrom);
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_home () 
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  curent = 1;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_end ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  curent=mp3dirsiz;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_pgup ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  curent-=YSIZ;
  if(curent<1) curent=1;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_pgdown ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  curent+=YSIZ;
  if(curent>mp3dirsiz) curent=mp3dirsiz;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_delete ()
{
  char buf[MAXPATHLEN];
  filesel_before ();

  if(mp3dir[curent].isdir) {
    sprintf (buf, "rm -rf \"%s\"", mp3dir[curent].name);
    //    log_debug (DEBUG_FILESEL, "filesel_delete: %s", buf);
    system (buf);
  }
  else if (delete_file(mp3dir[curent].name)) {
    readmp3dir();
    log_debug (DEBUG_FILESEL, "filesel_delete: mp3dirsiz= %d", mp3dirsiz);
    if(curent>mp3dirsiz) curent=mp3dirsiz;
    oldent=curent;
  }
  
  lcd_clear_screen ();
  
  /* Test if empty directory */
  if (mp3dirsiz > 0) {
    redrawall(curent,startfrom);
    filesel_update ();
  }
  else
    lcd_write_center (2, "No files!", ATTR_NORMAL);
}

void filesel_mark_file ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  if(mp3dir[curent].isdir==0)
    {
      mp3dir[curent].marked=1;
      sprintf (my_filename, "%s/%s", cdir, mp3dir[curent].name);
      omm_list_add (&my_list, my_filename);
      markchange=1;
    }
  if(curent<mp3dirsiz) curent++;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_unmark_file ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  if(mp3dir[curent].isdir==0)
    {
      mp3dir[curent].marked=0;
      sprintf (my_filename, "%s/%s", cdir, mp3dir[curent].name);
      omm_list_del (&my_list, my_filename);
      markchange=1;
    }
  if(curent<mp3dirsiz) curent++;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_toggle_file ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  if(mp3dir[curent].isdir==0)
    {
      mp3dir[curent].marked=1-mp3dir[curent].marked;
      markchange=1;

      sprintf (my_filename, "%s/%s", cdir, mp3dir[curent].name);
      if (mp3dir[curent].marked)
	omm_list_add (&my_list, my_filename);
      else
	omm_list_del (&my_list, my_filename);
    }
  if(curent<mp3dirsiz) curent++;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_mark_all_files ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  for(f=1;f<=mp3dirsiz;f++) if(mp3dir[f].isdir==0) {
    sprintf (my_filename, "%s/%s", cdir, mp3dir[f].name);
    omm_list_add (&my_list, my_filename);
    mp3dir[f].marked=1;
  }
  redrawall(curent,startfrom);
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_unmark_all_files ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  for(f=1;f<=mp3dirsiz;f++) if(mp3dir[f].isdir==0) {
    sprintf (my_filename, "%s/%s", cdir, mp3dir[f].name);
    omm_list_del (&my_list, my_filename);
    mp3dir[f].marked=0;
  }
  redrawall(curent,startfrom);
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_up ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  if(curent>1) curent--;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_down ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  if(curent<mp3dirsiz) curent++;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_right ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before ();
  curent-=YSIZ;
  if(curent<1) curent=1;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_left ()
{
  pthread_mutex_lock (&filename_mutex);
  filesel_before (); 
  curent+=YSIZ;
  if(curent>mp3dirsiz) curent=mp3dirsiz;
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_parent_dir ()
{
  if (!strcmp (cdir, mp3dir_base))
    return;

  pthread_mutex_lock (&filename_mutex);
  showbar(curent,0,startfrom);
  showmp3dir(startfrom,1);
  chdir("..");
  readmp3dir();
  showmp3dir(1,0); curent=oldent=startfrom=1;
  showbar(curent,1,startfrom);
  filesel_update ();
  pthread_mutex_unlock (&filename_mutex);
}

void filesel_enter_dir ()
{
  filesel_before ();

  if(mp3dir[curent].isdir)
    {
      pthread_mutex_lock (&filename_mutex);
      showbar(curent,0,startfrom);
      showmp3dir(startfrom,1);
      /* this may take a little while, so put cursor in sane place */
      lcd_move (fwinxpos(1)+1, fwinypos(1));
      chdir(mp3dir[curent].name);
      readmp3dir();
      if (mp3dirsiz ==0)
	chdir ("..");
      readmp3dir();
      showmp3dir(1,0); curent=oldent=startfrom=1;
      showbar(curent,1,startfrom);
      filesel_update ();
      pthread_mutex_unlock (&filename_mutex);

      return;
    }
}

void filesel_select ()
{
  filesel_before ();

  /*
  if(mp3dir[curent].isdir)
    return;
  */
#if 0
  if(mp3dir[curent].isdir)
    {
      pthread_mutex_lock (&filename_mutex);
      showbar(curent,0,startfrom);
      showmp3dir(startfrom,1);
      /* this may take a little while, so put cursor in sane place */
      lcd_move (fwinxpos(1)+1, fwinypos(1));
      chdir(mp3dir[curent].name);
      readmp3dir();
      if (mp3dirsiz ==0)
	chdir ("..");
      readmp3dir();
      showmp3dir(1,0); curent=oldent=startfrom=1;
      showbar(curent,1,startfrom);
      filesel_update ();
      pthread_mutex_unlock (&filename_mutex);

      return;
    }


  filesel_quit ();
#endif

  pthread_mutex_destroy (&filename_mutex);
  omm_api_close ();
  log_debug (DEBUG_FILESEL, "exiting OK.");


  if (mp3dir[curent].isdir) {
    if (fp_out != stderr)
      fclose (fp_out);

    if (the_outfile)
      my_unlink (the_outfile);
  }
  else {
    /* Marked files ? */
    if (my_list) {
      register omm_list *plist;

      for (plist = my_list ; plist ; plist = plist->next)
	fprintf (fp_out, "%s.%s\n", plist->str, mp3_suffix);

      /*      write (fd_fifo_reply, "0", 1);
	      exit (0);*/
    }
    else
      load_single_file (curent);

    if (fp_out != stderr)
      fclose (fp_out);
  }
	
  close_debug ();
  write (fd_fifo_reply, "0", 1);
  exit (0);
}
