#ident "$Id: fileselector.c,v 1.2 2000/12/12 10:29:40 pierre Exp pierre $"
/*
 * OMM File-selector
 */

#define EXTERN

#include "omm.h"
#include "fileselector.h"

static omm_list *my_list;
static boolean use_stdin = FALSE;
static char *outfile, *base_dir, *suffix;

extern int fd_fifo_reply;

/* States handling definition */
static omm_event_handler fileselector_event_handlers[] = {
  'd', OMM_STATE_INIT, lcd_dump_screen,		/* Dump screen to log file */
  'b', OMM_STATE_INIT, lcd_bg_screen,
  'r', OMM_STATE_INIT, lcd_fg_screen,
  12, OMM_STATE_INIT, filesel_redraw,		/* ^L redraw */
  'A'-0x40, OMM_STATE_INIT, filesel_home,		/* ^A home */
  'E'-0x40, OMM_STATE_INIT, filesel_end,		/* ^E end */
  'U'-0x40, OMM_STATE_INIT, filesel_pgup,		/* ^U PgUp */
  'V'-0x40, OMM_STATE_INIT, filesel_pgdown,		/* ^V PgDn */
  'j', OMM_STATE_INIT, filesel_down,		/* j Down */
  'k', OMM_STATE_INIT, filesel_up,		/* k Up */
  'h', OMM_STATE_INIT, filesel_right,		/* r Right */
  'l', OMM_STATE_INIT, filesel_left,		/* l Left */
  'a', OMM_STATE_INIT, filesel_parent_dir,	/* a Parent directory  */
  'n', OMM_STATE_INIT, filesel_enter_dir,	/* n Child directory */
  '\n', OMM_STATE_INIT, filesel_select,		/* LF Select file(s) */
  '\r', OMM_STATE_INIT, filesel_select,		/* CR Select file(s) */
  'D', OMM_STATE_INIT, filesel_delete,	        /* Remove a file */
  't', OMM_STATE_INIT, filesel_mark_file,	/* Mark a file */
  'u', OMM_STATE_INIT, filesel_mark_file,	/* Unmark a file */
  ' ', OMM_STATE_INIT, filesel_toggle_file,	/* toggle-mark file */
  'T', OMM_STATE_INIT, filesel_mark_all_files,	/* Mark all files */
  'U', OMM_STATE_INIT, filesel_unmark_all_files,/* Unmark all files */
  27, OMM_STATE_INIT, filesel_quit,
  0, 0, NULL
};

static void got_interrupt ()
{	
  log_debug (DEBUG_FILESEL, "Got interrupt!");
  exit (1);
}

static void usage (char *s)
{
  log_err ("Usage: %s [-s] [-d base_dir] [-S suffix] [-o outfile]\n", progname);
  exit (1);
}


/* Main function */
main (int ac, char **av)
{
  char *cp;

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  init_debug ();

  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {
      case 's' :
	use_stdin = TRUE; break;

      case 'd' :
	base_dir = *++av; break;

      case 'S' :
	suffix = *++av; break;

      case 'o' :
	outfile = *++av; break;

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }

  log_debug (DEBUG_FILESEL, "starting.");

  signal (SIGINT, got_interrupt);
  signal (SIGQUIT, got_interrupt);
  signal (SIGTERM, got_interrupt);

  /* Init API */
  omm_api_init (fileselector_event_handlers, FILESEL_FIFO, use_stdin);

  lcd_clear_screen ();

  /* Open filesel */
  filesel_init (suffix, base_dir, outfile);

  /* Main loop */
  omm_api_main_loop ();

  log_debug (DEBUG_FILESEL, "exiting QUIT.");
  close_debug ();

  if (outfile)
    my_unlink (outfile);

  write (fd_fifo_reply, "1", 1);
  exit (1);
}
