#ident "$Id: manager.c,v 1.3 2000/12/30 22:21:27 pierre Exp pierre $"
/*
 * OMM Manager
 */

#define EXTERN

#include "omm.h"
#include "manager.h"

static boolean use_stdin = FALSE;
static char *current_fifo = MANAGER_FIFO;
static char *old_fifo;
static char *omm_dir = DEFAULT_OMMDIR;
static int old_state = OMM_STATE_INIT;
static boolean playlist_mode = FALSE;
static boolean napster_started = FALSE;

/* States handling definition */
static omm_event_handler manager_event_handlers[] = {
  'c', OMM_STATE_INIT, manager_start_cdplayer,
  'c', MANAGER_STATE_MP3, manager_start_cdplayer,
  'c', MANAGER_STATE_ENC, manager_start_cdplayer,
  'c', MANAGER_STATE_NAPSTER, manager_start_cdplayer,
  'z', OMM_STATE_INIT, manager_start_encoder,
  'z', MANAGER_STATE_CD, manager_start_encoder,
  'z', MANAGER_STATE_MP3, manager_start_encoder,
  'z', MANAGER_STATE_NAPSTER, manager_start_encoder,
  'm', OMM_STATE_INIT, manager_start_mp3player,
  'm', MANAGER_STATE_CD, manager_start_mp3player,
  'm', MANAGER_STATE_ENC, manager_start_mp3player,
  'm', MANAGER_STATE_NAPSTER, manager_start_mp3player,
  'z', MANAGER_STATE_MP3, manager_start_encoder,
  'm', MANAGER_STATE_MP3, manager_restart_mp3player,
  'M', OMM_STATE_INIT, manager_toggle_playlist_mode,
  'M', MANAGER_STATE_CD, manager_toggle_playlist_mode,
  'M', MANAGER_STATE_MP3, manager_toggle_playlist_mode,
  'M', MANAGER_STATE_FILESEL, manager_toggle_playlist_mode,
  'f', MANAGER_STATE_CD, manager_start_fileselector,
  'f', MANAGER_STATE_MP3, manager_start_fileselector,
  'f', MANAGER_STATE_ENC, manager_start_fileselector,
  'e', MANAGER_STATE_ENC, manager_start_enc_editor,
  'n', OMM_STATE_INIT, manager_start_napster,
  'n', MANAGER_STATE_CD, manager_start_napster,
  'n', MANAGER_STATE_MP3, manager_start_napster,
  'n', MANAGER_STATE_ENC, manager_start_napster,
  '\n', MANAGER_STATE_FILESEL, manager_close_fileselector,
  27, MANAGER_STATE_FILESEL, manager_quit_fileselector,
  27, MANAGER_STATE_NAPSTER, manager_quit_napster,
  27, OMM_STATE_INIT, manager_quit,
  27, MANAGER_STATE_CD, manager_void,
  27, MANAGER_STATE_MP3, manager_void,
  'q', MANAGER_STATE_FILESEL, manager_quit,
  'x', OMM_STATE_INIT, manager_quit,
  'x', MANAGER_STATE_CD, manager_quit,
  'x', MANAGER_STATE_MP3, manager_quit,
  'x', MANAGER_STATE_ENC, manager_quit,
  'x', MANAGER_STATE_NAPSTER, manager_quit,
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, manager_default_handler
};

/* Got a signal */
static void got_interrupt (int sig)
{	
  log_debug (DEBUG_FILESEL, "Got interrupt %d!", sig);
  exit (1);
}

#if 0

/* Moved to write_fifo() */

/* Read from reply fifo */
static int read_fifo_reply (char *fifo)
{
  char c, buf[MAXPATHLEN];
  int r, fd;

  sprintf (buf, "%s_reply", fifo);

  fd = open (buf, O_RDWR);
  if (fd > 0) {
    r = read (fd, &c, 1);
    log_debug (DEBUG_MANAGER, "read_fifo: got %c [%02x]", c, c&255);
    close (fd);
    return r;
  }
  else {
    log_err ("read_fifo_reply: %s: %s", buf, errno);
    return -1;
  }
}
#endif

/* 
 * Write to application fifo, may wait a reply 
 */
static int write_fifo (char *fifo, char cmd, boolean wait_reply)
{
  int fd = open (fifo, O_RDWR), r;
  char c, buf[MAXPATHLEN];

  if (fd > 0) {
    log_debug (DEBUG_MANAGER, "write_fifo: %02x (%c) to %s", cmd&255, cmd, fifo);
    if (write (fd, &cmd, 1) < 0) {
      close (fd);
      log_err ("write_fifo: write error: %s", strerror (errno));
      return -1;
    }
  }
  else {
    log_err ("write_fifo: open error: %s", strerror (errno));
    return -1;
  }

  close (fd);

  if (wait_reply) {
    sprintf (buf, "%s_reply", fifo);
    if ((fd = open (buf, O_RDWR)) > 0) {
      r = read (fd, &c, 1);
      log_debug (DEBUG_MANAGER, "write_fifo: got reply %c [%02x]", c, c&255);
      close (fd);
      return r;
    }
    else {
      log_err ("write_fifo: %s: %s", buf, errno);
      return -1;
    }
  }
  else
    return 0;
}

/* Write to the current fifo (foreground app) */
static int write_current_fifo (int cmd, boolean wait_reply)
{
  return write_fifo (current_fifo, cmd, wait_reply);
}

/* Execute an application */
static void manager_execute (char *prog)
{
 char buf[MAXPATHLEN];

 log_debug (DEBUG_MANAGER, "EXECUTE %s ***", prog);

 sprintf (buf, "%s/%s/%s &", omm_dir, BIN_DIR, prog);
 system (buf);
}

/* Just eat the char a do nothing */
static void manager_void ()
{
  omm_api_event ev = omm_api_get_current_event();
  log_debug (DEBUG_MANAGER, "manager_void: %c ignored!", ev);
}

/*
 * Default input handler: send the char to the current application
 */
static void manager_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  if (ev >= ' ')
    log_debug (DEBUG_MANAGER, "manager_default_handler: '%c' to %s", ev, current_fifo);
  else
    log_debug (DEBUG_MANAGER, "manager_default_handler: '%02x' to %s", ev&255, current_fifo);

  write_current_fifo (ev, FALSE);
}

/* 
 * Start applications 
 */
static void manager_start_cdplayer ()
{
  if (omm_state != OMM_STATE_INIT && omm_state != MANAGER_STATE_NAPSTER)
    write_current_fifo ('q', FALSE);

  if (omm_state == MANAGER_STATE_NAPSTER)
    write_current_fifo ('B', TRUE);

  omm_state = MANAGER_STATE_CD;
  current_fifo = CDPLAYER_FIFO;

  manager_execute ("cdplayer_x11");
}

static void manager_start_mp3player ()
{
  char buf[MAXPATHLEN], filename[MAXPATHLEN];
  struct stat statb;
  FILE *fp;

  log_debug (DEBUG_MANAGER, "manager_start_mp3player: starting");

  if (omm_state != OMM_STATE_INIT && omm_state != MANAGER_STATE_NAPSTER)
    write_current_fifo ('q', FALSE);

  if (omm_state == MANAGER_STATE_NAPSTER)
    write_current_fifo ('B', TRUE);

  omm_state = MANAGER_STATE_MP3;
  current_fifo = MP3PLAYER_FIFO;

  /* Check for list */
  if (playlist_mode) {
    sprintf (filename, "%s/%s/%s.%s", omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, METALIST_SUFFIX);
    if (stat (filename, &statb) < 0) {
      sprintf (buf, "mp3player_x11 -d %s/%s", omm_dir, PLAYLISTS_DIR);
    }
    else {
      /* Read the list of playlists */
      sprintf (buf, "mp3player_x11 -d %s/%s -p ", omm_dir, PLAYLISTS_DIR);

      fp = fopen (filename, "r");
      if (fp) {
	while (fgets (filename, sizeof(filename), fp)) {
	  filename[strlen(filename)-1] = 0;
	  strcat (buf, basename(filename));
	  strcat (buf, " ");
	}

	fclose (fp);
      }
    }
  }
  /* Load a single play-list */
  else {
    sprintf (filename, "%s/%s/%s.%s", omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, PLAYLIST_SUFFIX);
    if (stat (filename, &statb) < 0) {
      log_debug (DEBUG_MANAGER, "manager_start_mp3player: nothing to play!");
      //    omm_state = OMM_STATE_INIT;
      //    lcd_clear_screen ();
      //    lcd_write_center (1, "Nothing to play !", ATTR_BLINK);
      sprintf (buf, "mp3player_x11 -d %s/%s", omm_dir, PLAYLISTS_DIR);
    }
    else
      sprintf (buf, "mp3player_x11 -d %s/%s -p %s.%s", omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, PLAYLIST_SUFFIX); 
  }

  /*
    omm_state = MANAGER_STATE_FILESEL;
    current_fifo = FILESEL_FIFO;
    sprintf (buf, "fileselector_x11 2> %s", playlist); 
  */

  manager_execute (buf);
}

static void manager_start_encoder ()
{
  if (omm_state != OMM_STATE_INIT && omm_state != MANAGER_STATE_NAPSTER)
    write_current_fifo ('q', FALSE);

  if (omm_state == MANAGER_STATE_NAPSTER)
    write_current_fifo ('B', TRUE);

  omm_state = MANAGER_STATE_ENC;
  current_fifo = ENCODER_FIFO;

  manager_execute ("encoder_x11");
}

static void manager_start_napster ()
{
  /* Set current app to background state */
  if (omm_state != OMM_STATE_INIT) {
    old_fifo = current_fifo;
    old_state = omm_state;
    write_current_fifo ('B', TRUE);
    //    sleep (1); /* should use fd_fifo_reply instead! */
  }

  omm_state = MANAGER_STATE_NAPSTER;
  current_fifo = NAPSTER_FIFO;

  if (!napster_started) {
    manager_execute ("napster_x11");
    napster_started = TRUE;
  }
  else {
    write_current_fifo ('F', TRUE);
  }
}

static void manager_start_enc_editor ()
{
  char buf[MAXPATHLEN];

  /* Set current app to background state */
  if (omm_state != OMM_STATE_INIT) {
    old_fifo = current_fifo;
    old_state = omm_state;
    write_current_fifo ('B', TRUE);
  }

  /* Open file-selector and save selection in a temp file */
  omm_state = MANAGER_STATE_FILESEL;
  current_fifo = FILESEL_FIFO;

  sprintf (buf, "fileselector_x11 -d %s/%s/%s -S %s", omm_dir, SPOOL_DIR, EXTRACT_DIR, TRACK_SUFFIX);
  manager_execute (buf);
}


static void manager_restart_mp3player ()
{
  char buf[MAXPATHLEN], *filename;
  register int i;
  FILE *fp;

  log_debug (DEBUG_MANAGER, "manager_restart_mp3player: starting");


  if (playlist_mode) {
    sprintf (buf, "%s/%s/%s.%s", omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, METALIST_SUFFIX); 
    if ((fp = fopen (buf, "r"))) {
      write_current_fifo ('r', FALSE);

      while (fgets (buf, sizeof(buf), fp)) {
	filename = basename(buf);
	for (i = 0 ; buf[i] ; i++)
	  write_current_fifo (filename[i], FALSE);
      }

      write_current_fifo ('\004', FALSE);
      fclose (fp);
    }
  }
  else {
    //    sprintf (buf, "%s/%s/%s.%s", omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, MP3_SUFFIX); 
    sprintf (buf, "%s.%s", DEFAULT_PREFIX, PLAYLIST_SUFFIX);
    write_current_fifo ('r', FALSE);
    for (i = 0 ; buf[i] ; i++)
      write_current_fifo (buf[i], FALSE);
    write_current_fifo ('\n', FALSE);
    write_current_fifo ('\004', FALSE);
  }

  log_debug (DEBUG_MANAGER, "manager_restart_mp3player: exiting");
}



/* Open file-selector */
static void do_start_fileselector ()
{
  char buf[MAXPATHLEN];

  if (playlist_mode)
    sprintf (buf, "fileselector_x11 -d %s/%s -S %s -o /tmp/%s.%s", omm_dir, PLAYLISTS_DIR, PLAYLIST_SUFFIX, DEFAULT_PREFIX, METALIST_SUFFIX);
  else
    sprintf (buf, "fileselector_x11 -d %s/%s -S %s -o /tmp/%s.%s", omm_dir, MP3_DIR, MP3_SUFFIX, DEFAULT_PREFIX, PLAYLIST_SUFFIX);

  manager_execute (buf);
}

static void manager_start_fileselector ()
{
  char buf[MAXPATHLEN];

  /* Set current app to background state */
  if (omm_state != OMM_STATE_INIT) {
    old_fifo = current_fifo;
    old_state = omm_state;
    write_current_fifo ('B', TRUE);
  }

  /* Open file-selector and save selection in a temp file */
  omm_state = MANAGER_STATE_FILESEL;
  current_fifo = FILESEL_FIFO;
  do_start_fileselector ();
}


/*
 * Quit applications
 */

static void manager_quit_napster ()
{
  /* Send quit command and wait for reply */
  write_current_fifo ('q', FALSE);
  napster_started = FALSE;

  /* Back to background app */
  if (old_state != OMM_STATE_INIT) {
    current_fifo = old_fifo;
    omm_state = old_state;

    /* Set to foreground */
    write_current_fifo ('F', FALSE);
    old_state = OMM_STATE_INIT;
  }
}


/* Close file-selector with correct selection */
static void manager_close_fileselector ()
{
  char buf[MAXPATHLEN], playlist_name[MAXPATHLEN];
  struct stat statb;

  /* Send select command and wait for reply */
  log_debug (DEBUG_MANAGER, "manager_close_fileselector: send \\n");
  write_current_fifo ('\n', TRUE);
  log_debug (DEBUG_MANAGER, "manager_close_fileselector: read reply");
  //  read_fifo_reply (current_fifo);
  log_debug (DEBUG_MANAGER, "manager_close_fileselector: OK");

  /* Back to background app */
  if (old_state != OMM_STATE_INIT) {
    current_fifo = old_fifo;
    omm_state = old_state;
    write_current_fifo ('F', TRUE);
    old_state = OMM_STATE_INIT;
  }

  /* Copy tmp file to default playlist */
  if (playlist_mode) {
    sprintf (buf, "/tmp/%s.%s", DEFAULT_PREFIX, METALIST_SUFFIX);
    if (!stat (buf, &statb)) {
      sprintf (buf, "cp /tmp/%s.%s %s/%s/%s.%s", DEFAULT_PREFIX, METALIST_SUFFIX, omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, METALIST_SUFFIX);
      manager_execute (buf);
    }
  }
  else {
    sprintf (buf, "/tmp/%s.%s", DEFAULT_PREFIX, PLAYLIST_SUFFIX);
    if (!stat (buf, &statb)) {
      sprintf (buf, "cp /tmp/%s.%s %s/%s/%s.%s", DEFAULT_PREFIX, PLAYLIST_SUFFIX, omm_dir, PLAYLISTS_DIR, DEFAULT_PREFIX, PLAYLIST_SUFFIX);
      manager_execute (buf);
    }
  }

}

/* Quit file-selector without selecting files */
static void manager_quit_fileselector ()
{
  /* Send quit command and wait for reply */
  write_current_fifo ('\x1b', TRUE);
  //  read_fifo_reply (current_fifo);

  /* Back to background app */
  if (old_state != OMM_STATE_INIT) {
    current_fifo = old_fifo;
    omm_state = old_state;

    /* Set to foreground */
    write_current_fifo ('F', TRUE);
    old_state = OMM_STATE_INIT;
  }
}


/* Toggle dir/playlist mode */
static void manager_toggle_playlist_mode ()
{
  if (playlist_mode)
    playlist_mode = FALSE;
  else
    playlist_mode = TRUE;

  log_debug (DEBUG_MANAGER, "playlist_mode= %d", playlist_mode);

  /* If running fileselector stop it and start again */
  if (omm_state == MANAGER_STATE_FILESEL) {
    write_current_fifo ('\x1b', TRUE);
    //    read_fifo_reply (current_fifo);
    do_start_fileselector ();
  }
}

/* Quit manager */
static void manager_quit ()
{
  /* Quit current application and exit */
  if (omm_state != OMM_STATE_INIT)
    write_current_fifo ('q', FALSE);

  omm_state = OMM_STATE_INIT;
  omm_api_close ();
}

/* write a string scolling from the right border to x */
static void write_scrolled (int x, int y, char *s)
{
  register int xx, i = 0;
  char buf[2];

  buf[1] = 0;

  for (i = 0 ; i < strlen(s) ; i++) {
    buf[0] = *(s+i);
    for (xx = LCD_WIDTH-1 ; xx >= x+i ; xx--) {
      lcd_write (xx+i, y, buf, ATTR_NORMAL);
      usleep (50000);
      lcd_write (xx+i, y, " ", ATTR_NORMAL);
      usleep (50000);
    }

    lcd_write (x+i, y, buf, ATTR_NORMAL);
  }
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s [-s]\n", s);
  exit (1);
}


/* Main function */
main (int ac, char **av)
{
  char *cp, *new_omm_dir, buf[256];
  unsigned long my_win;

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if ((new_omm_dir = getenv ("OMMDIR")))
    omm_dir = new_omm_dir;

  init_debug ();

  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {
      case 's' :
	use_stdin = TRUE; break;

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }


  log_debug (DEBUG_MANAGER, "starting.");

  signal (SIGINT, got_interrupt);
  signal (SIGQUIT, got_interrupt);
  signal (SIGTERM, got_interrupt);

  /* Init API */
  omm_api_init (manager_event_handlers, MANAGER_FIFO, use_stdin);
  my_win = lcd_get_window ();

  if (my_win) {
    sprintf (buf, "OMMWIN=%x", my_win);
    putenv (buf);
  }
    

  lcd_clear_screen ();
  lcd_write_center (1, "Welcome to    ", ATTR_NORMAL);
  write_scrolled (14, 1, "OMM");


  /* Main loop */
  omm_api_main_loop ();

  log_debug (DEBUG_MANAGER, "exiting.");
  close_debug ();
}
