#ident "$Id: mp3player.c,v 1.5 2000/12/12 10:29:59 pierre Exp pierre $"
/*
 * OMM MP3 player
 */

#define EXTERN

#include "omm.h"
#include "mp3player.h"

static int max_track;	/* Max # of tracks, starting from 1 */
static int current_track = 1, old_current_track;
static int old_minutes = -1, old_seconds = -1, old_mp3_mode = -1;
static int pipe_in[2], pipe_out[2];
static pthread_t play_track_thread, display_info_thread;
static char mp3_end;
static int mp3_mode;
static int mpg123_pid;
static pthread_mutex_t mpg123_mutex;
static pthread_mutex_t track_mutex;
static int write_cnt, read_cnt;
static mp3_track *tracks[MAXTRACKS];
static int display_type = DISPLAY_TRACK_LENGTH;
static float sectotal;
static char *mode_str[] = { "STOP", "PLAY", "PAUSE" };
static float sec;
static char *playlist_file, *playlist[MAXTRACKS], **playlist_list;
static char *playlist_dir;
static boolean use_stdin = FALSE;
static boolean reloading = FALSE;
static boolean info_process_running = FALSE;

/* States definition */
static omm_event_handler mp3_event_handlers[] = {
  'd', OMM_STATE_INIT, lcd_dump_screen,		/* Dump screen to log file */
  'B', OMM_STATE_INIT, lcd_bg_screen,
  'F', OMM_STATE_INIT, lcd_fg_screen,
  'h', OMM_STATE_INIT, mp3_player_pause,	 /* Pause */
  'p', OMM_STATE_INIT, mp3_player_play,		 /* Play */
  '+', OMM_STATE_INIT, mp3_player_next_track,	 /* Track + */
  '-', OMM_STATE_INIT, mp3_player_previous_track,/* Track - */
  'q', OMM_STATE_INIT, mp3_player_quit,		 /* Quit */
  's', OMM_STATE_INIT, mp3_player_stop,		 /* Stop */
  't', OMM_STATE_INIT, mp3_player_toggle_display,/* Toggle display: time/title */
  'r', OMM_STATE_INIT, mp3_player_reload,	 /* Reload */
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, mp3_player_default_handler,
  0, 0, NULL
};


static void got_interrupt (int sig)
{	
  log_debug (DEBUG_MP3PLAYER, "Got interrupt %d!", sig);
  mp3_player_quit ();  
  exit (1);
}


static void write_mpg123_cmd (char *s, int size)
{
  //  log_debug (DEBUG_MP3PLAYER, "write_mpg123_cmd: [%s] %d bytes", s, size);

  if (write (PLAYER_OUT, s, size) < 0) {
    log_err ("write_mpg123_cmd: %s: %s", s, strerror(errno));
    mp3_player_quit ();
  }

  write_cnt++;
}

static void mp3_player_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  if (ev >= ' ')
    log_debug (DEBUG_MP3PLAYER, "mp3_player_default_handler: got '%c' in state %02x", ev, omm_state);
  else
    log_debug (DEBUG_MP3PLAYER, "mp3_player_default_handler: got '%02x' in state %02x", ev, omm_state);
}

static void read_mpg123_status (char *s, int size)
{
  char c = 0;
  int cnt = 0;

  //  log_debug (DEBUG_MP3PLAYER, "read_mpg123_status: %d bytes", size);

  memset ((void*)s, 0, size);
  while (c != '\n' && cnt < size) {
    if (read (PLAYER_IN, &c, 1) < 0) {
      log_err ("read_mpg123_status: %s: %s", s, strerror(errno));
      mp3_player_quit ();
    }
    
    //    log_debug (DEBUG_MP3PLAYER, "c= %02x %c", c&255, c);

    if (c != '\n') {
      *(s+cnt) = c;
      cnt++;
    }
  }
  
  read_cnt++;
}


static void display_track_info ()
{
  char buf[LCD_WIDTH+1];
  mp3_track *the_track = tracks[current_track-1];

  if (!the_track) {
    log_err ("track %d not allocated!", current_track);
    return;
  }

  sprintf (buf, "%3s %1d %-3d %.1f %3s", the_track->type, the_track->layer,  the_track->bitrate, (float)(the_track->frequency)/1000, the_track->mode);
  lcd_write (2, 0, buf, ATTR_NORMAL);
}

static void display_zero_time ()
{
  char buf[LCD_WIDTH+1];

  log_debug (DEBUG_MP3PLAYER, "display_zero_time");
  
  lcd_init_hscroll ();
  sprintf (buf, "Track   00:00 %02d", current_track);
  lcd_write_center (1, buf, ATTR_NORMAL);
}



static void display_track_length ()
{
  char buf[LCD_WIDTH+1];
  mp3_track *the_track = tracks[current_track-1];

  if (display_type != DISPLAY_TRACK_LENGTH)
    return;

  if (!the_track) {
    log_err ("track %d not allocated!", current_track);
    return;
  }

  sprintf (buf, "Length   %02d:%02d   ", (unsigned int)(the_track->sectotal)/60, (unsigned int)(the_track->sectotal)%60);
  lcd_write_center (2, buf, ATTR_NORMAL);
}

static void display_total_length ()
{
  char buf[LCD_WIDTH+1];

  if (display_type != DISPLAY_TOTAL_LENGTH)
    return;

  sprintf (buf, "Total   %02d:%02d %02d", (unsigned int)sectotal/60, (unsigned int)sectotal%60, max_track);
  lcd_write_center (2, buf, ATTR_NORMAL);
}


static void copy_and_cut (char *dst, char *src, int maxlen)
{
  register char *p;

  strncpy (dst, src, maxlen);
  p = dst+maxlen-1;
  while (*p == ' ' && p != dst)
    p--;
  
  if (p != dst)
    p++;

  *p = 0;
}

static void play_track (int track)
{
  char sendbuf[MAXPATHLEN], replybuf[256];
  mp3_track *the_track = tracks[track-1];

  if (!the_track) {
    log_err ("play_track: track %d not allocated!", track);
    return;
  }
    
  pthread_mutex_lock (&mpg123_mutex);
  sprintf (sendbuf, "l %s", the_track->pathname);
  write_mpg123_cmd (sendbuf, strlen(sendbuf));
  read_mpg123_status (replybuf, sizeof(replybuf));
  sscanf (&replybuf[3], "%d", &mp3_mode);


  log_debug (DEBUG_CDPLAYER, "play_track mode= %d [%s]", mp3_mode, replybuf);

  write_mpg123_cmd ("R", 1);
  read_mpg123_status (replybuf, sizeof(replybuf));
  sscanf (&replybuf[3], "%d", &mp3_mode);
  pthread_mutex_unlock (&mpg123_mutex);
}

/* Free the tracks tab from start to end */
void free_tracks (int start, int end)
{
  register int i;

  if (end > MAXTRACKS-1)
    end = MAXTRACKS-1;

  for (i = start ; i <= end ; i++) {
    if (tracks[i]) {
      if (tracks[i]->tag)
	free (tracks[i]->tag);
      if (tracks[i]->pathname)
	free (tracks[i]->pathname);
      if (tracks[i]->filename)
	free (tracks[i]->filename);

      free (tracks[i]);
    }
  }
  
  log_debug (DEBUG_MP3PLAYER, "free_tracks: %d track(s) freed", end-start+1);
  sectotal = 0.0;
}

/* Load a list of MP3 files in the tracks tab */
int load_files (char **list, int start)
{
  register int i = start, j = start;
  char sendbuf[MAXPATHLEN], replybuf[256];
  int dummy;
  float sec;
  mp3_track *the_track;
  struct stat statb;

  sectotal = 0.0;

  while (list[i]) {
    if (stat (list[i], &statb) < 0) {
      log_err ("load_files: %s: %s", list[i], strerror(errno));
      free (list[i]); list[i] = NULL;
      i++;
      continue;
    }

    log_debug (DEBUG_MP3PLAYER, "loading track %d: %s", i, list[i]);
    if (!(tracks[j] = the_track = (mp3_track *) calloc (1, sizeof(mp3_track)))) {
      log_err ("Can't allocate track[%d] !", j);
      mp3_player_quit ();
    }
      
    if (!(the_track->pathname = strdup (list[i]))) {
      log_err ("Can't allocate pathname[%d] !", j);
      mp3_player_quit ();
    }

    sprintf (sendbuf, "i %s", the_track->pathname);
    write_mpg123_cmd (sendbuf, strlen(sendbuf));

    /* Get ID3 or filename */
    read_mpg123_status (replybuf, sizeof(replybuf));

    if (replybuf[1] != 'I') {
      log_err ("Synchro error, %c instead of I !");
      mp3_player_quit ();
    }

    /* ID3 ? */
    if (!strncmp (&replybuf[3], "ID3:", 4)) {
      struct id3tag *tag = (struct id3tag *) &replybuf[7];
	
      if (!(the_track->tag = (struct id3tag *)calloc (1, sizeof(struct id3tag)))) {
	log_err ("Can't alloc ID3 tags[%d] !", j);
	mp3_player_quit ();
      }

      copy_and_cut (the_track->tag->title, tag->title, 30);
      copy_and_cut (the_track->tag->artist, tag->artist, 30);
      copy_and_cut (the_track->tag->album, tag->album, 30);
      copy_and_cut (the_track->tag->year, tag->year, 4);
      copy_and_cut (the_track->tag->comment, tag->comment, 30);
      copy_and_cut (the_track->tag->genre, tag->genre, 30);

      if (!(the_track->filename = strdup(basename(the_track->pathname)))) {
	log_err ("Can't alloc filename[%d] !", j);
	mp3_player_quit ();
      }

      remove_suffix (the_track->filename, ".mp3");
    }
    else {
      if (!(the_track->filename = strdup(&replybuf[3]))) {
	log_err ("Can't alloc filename[%d] !", j);
	mp3_player_quit ();
      }
    }

    /* Track info */
    read_mpg123_status (replybuf, sizeof(replybuf));

    if (replybuf[1] != 'S') {
      log_err ("Synchro error, %c instead of S !");
      mp3_player_quit ();
    }

    sscanf (&replybuf[3], "%4s %d %ld %4s %d %d %d %d %d %d %d %d", &(the_track->type), &(the_track->layer), &(the_track->frequency), &(the_track->mode), &dummy, &dummy, &dummy, &dummy, &dummy, &dummy, &(the_track->bitrate), &dummy);
    log_debug (DEBUG_MP3PLAYER, "type= %s layer= %d freq= %d, mod= %s, bitrate= %d", the_track->type, the_track->layer, the_track->frequency, the_track->mode, the_track->bitrate);

    /* 1rst track info */
    read_mpg123_status (replybuf, sizeof(replybuf));

    if (replybuf[1] != 'F') {
      log_err ("Synchro error, %c instead of F !");
      mp3_player_quit ();
    }

    sscanf (&replybuf[3], "%d %d %f %f", &dummy, &dummy, &sec, &(the_track->sectotal));
    sectotal += the_track->sectotal;

    log_debug (DEBUG_MP3PLAYER, "OK!");

    free (list[i]); list[i] = NULL;
    i++; 
    j++;
  }
  
  return j;
}

int load_play_list (char *name, char **list, int start)
{
  char filename[MAXPATHLEN], playlist_path[MAXPATHLEN];
  FILE *fp;
  register int i = start;

  if (i >= MAXTRACKS)
    return 0;

  if (playlist_dir)
    sprintf (playlist_path, "%s/%s", playlist_dir, name);
  else
    strcpy (playlist_path, name);

  if (!(fp = fopen (playlist_path, "r"))) {
    log_err ("load_play_list: %s: %s", name, strerror(errno));
    return 0;
  }

  while (fgets (filename, sizeof(filename), fp) && i < MAXTRACKS)  {
    filename[strlen(filename)-1] = 0;
    list[i] = strdup (filename);
    i++;
  }

  return load_files (list, start);
}

static void *display_info_process (void *dummy)
{
  mp3_track *the_track;
  struct id3tag *tag;
  int minutes, seconds;
  char buf[LCD_WIDTH+1];
  struct timeval tv;
  struct timezone tz;
  unsigned long tms, tms_old, starting_time;

  log_debug (DEBUG_CDPLAYER, "display_info_process: starting");
  info_process_running = TRUE;

  tms = tms_old = 0;
  starting_time = time (0);
  old_minutes = -1;
  old_seconds = -1;
  old_mp3_mode = -1;
  old_current_track = -1;

  while (!reloading) {
    /* Display time */
    minutes = (unsigned int)sec/60;
    seconds = (unsigned int)sec%60;

    /* Optimize the display: update modified values */
    if (old_minutes != minutes) {
      sprintf (buf, "%02d", minutes);
      lcd_write (10, 1, buf, ATTR_NORMAL);
    }

    if (old_seconds != seconds) {
      /*
      gettimeofday (&tv, &tz);
      tms = (tv.tv_sec - starting_time) * 1000 + tv.tv_usec / 1000;
      log_debug (DEBUG_MP3PLAYER, "%ld %ld %ld", tms_old, tms, tms-tms_old);
      if (tms_old && ((tms-tms_old) < 1000)) {
	log_debug (DEBUG_MP3PLAYER, "attente %ld %ld!", tms-tms_old, 1000 * (1000 - (tms-tms_old)));
	usleep (1000 * (1000 - (tms-tms_old)));
      }

      tms_old = tms;
      */

      sprintf (buf, "%02d", seconds);
      lcd_write (13, 1, buf, ATTR_NORMAL);
    }


    pthread_mutex_lock (&track_mutex);
    if (old_current_track != current_track) {
      sprintf (buf, "%02d", current_track);
      lcd_write (16, 1, buf, ATTR_NORMAL);
      lcd_clear_line (0);
      display_track_info ();
      display_track_length ();
      lcd_init_hscroll ();
    }
    pthread_mutex_unlock (&track_mutex);

    if (old_mp3_mode != mp3_mode) {
      if (mp3_mode == MODE_STOPPED) {
	display_zero_time ();
	display_track_length ();
	lcd_init_hscroll ();
      }

      log_debug (DEBUG_MP3PLAYER, "mp3_mode = %d", mp3_mode);      
      lcd_write_center (3, mode_str[mp3_mode], ATTR_NORMAL);
    }

    old_minutes = minutes;
    old_seconds = seconds;
    old_mp3_mode = mp3_mode;

    pthread_mutex_lock (&track_mutex);
    old_current_track = current_track;
    pthread_mutex_unlock (&track_mutex);

    the_track = tracks[current_track-1];
    tag = the_track->tag;

    switch (display_type) {

    case DISPLAY_TITLE:
      lcd_write_hscroll (-1, 2, tag->title);
      break;

    case DISPLAY_ARTIST:
      lcd_write_hscroll (-1, 2, tag->artist);
      break;

    case DISPLAY_ALBUM:
      lcd_write_hscroll (-1, 2, tag->album);
      break;

    case DISPLAY_YEAR:
      lcd_write_hscroll (-1, 2, tag->year);
      break;

    case DISPLAY_COMMENT:
      lcd_write_hscroll (-1, 2, tag->comment);
      break;

    case DISPLAY_GENRE:
      lcd_write_hscroll (-1, 2, tag->genre);
      break;

    case DISPLAY_FILENAME:
      lcd_write_hscroll (-1, 2, the_track->filename);
      break;
    
    default: break;
    }

    usleep (500000);
  }

  log_debug (DEBUG_MP3PLAYER, "display_info_process: exiting");
  lcd_clear_screen ();
  info_process_running = FALSE;
  pthread_exit (0);
}

static void *play_track_process (void *dummy)
{
  char replybuf[256], buf[LCD_WIDTH+1];
  int fcount, fleft;
  float secleft;
  mp3_track *the_track;
  struct id3tag *tag;

  log_debug (DEBUG_MP3PLAYER, "===> play_track_process: start\n");

  for (;;) {
    pthread_mutex_lock (&mpg123_mutex);

    /* Test mp3_end here in case of slow display who locked the mutex! */
    if (mp3_end) {
      pthread_mutex_unlock (&mpg123_mutex);
      break;
    }

    write_mpg123_cmd ("F", 1);
    read_mpg123_status (replybuf, sizeof(replybuf));
    sscanf (&replybuf[3], "%d %d %f %f %d", &fcount, &fleft, &sec, &secleft, &mp3_mode);
    pthread_mutex_unlock (&mpg123_mutex);
    
    /* End of current track */
    if (mp3_mode == MODE_STOPPED && !mp3_end) {
      log_debug (DEBUG_MP3PLAYER, "end of track cur= %d old=%d! mode= %d end= %d", current_track, old_current_track, mp3_mode, mp3_end);

      /* Try next track */
      if (current_track == max_track) {
	log_debug (DEBUG_MP3PLAYER, "max track!");
	mp3_end = 1;
      }
      else {
	/* Play next track */
	pthread_mutex_lock (&track_mutex);
	//	next_track ();
	current_track++;
	pthread_mutex_unlock (&track_mutex);
	play_track (current_track);
      }
    }

    //    log_debug (DEBUG_MP3PLAYER, "play_track_process: mode= %d end= %d", mp3_mode, mp3_end);
    usleep (500000);
  }

  log_debug (DEBUG_MP3PLAYER, "<=== play_track_process: exit, mode= %d", mp3_mode);
  pthread_exit (0);
}

/* Play from the current track */
static void mp3_player_play ()
{
  char buf[LCD_WIDTH+1];

  if (mp3_mode == MODE_PLAYING)
    return;

  mp3_end = 0;

  /* If paused, just resume */
  if (mp3_mode == MODE_PAUSED) {
    mp3_player_pause ();
  }
  else {
    play_track (current_track);

    /* Start time displaying thread */
    if (pthread_create (&play_track_thread, &thread_attr, play_track_process, NULL) != 0)
      log_err ("pthread_create error for play_track_process");
  }

  /* Force  time+track redraw */
  old_minutes = old_seconds = old_mp3_mode = -1;
}

static void do_stop ()
{
  char replybuf[256];

  log_debug (DEBUG_MP3PLAYER, "do_stop!");

  if (mp3_mode == MODE_STOPPED)
    return;

  pthread_mutex_lock (&mpg123_mutex);
  write_mpg123_cmd ("S", 1);
  read_mpg123_status (replybuf, sizeof(replybuf));
  sscanf (&replybuf[3], "%d", &mp3_mode);
  pthread_mutex_unlock (&mpg123_mutex);
}

static void mp3_player_stop ()
{
  //  current_track = 1
  mp3_end = 1;
  do_stop ();

}

static void mp3_player_pause ()
{
  char replybuf[256];

  pthread_mutex_lock (&mpg123_mutex);
  write_mpg123_cmd ("P", 1);
  read_mpg123_status (replybuf, sizeof(replybuf));
  sscanf (&replybuf[3], "%d", &mp3_mode);
  pthread_mutex_unlock (&mpg123_mutex);
}

static void mp3_player_quit ()
{
  mp3_end = 1;
  if (mpg123_pid) {
    log_debug (DEBUG_MP3PLAYER, "Killing '%s' process %d!", MPG123, mpg123_pid);
    kill (mpg123_pid, SIGTERM);
  }

  omm_state = OMM_STATE_INIT;
  omm_api_close ();
}

static void mp3_player_next_track ()
{
  int old_mode = mp3_mode;

  if (current_track < max_track) {
    if (mp3_mode != MODE_STOPPED) {
      if (mp3_mode == MODE_PAUSED)
	mp3_end = 1;
      do_stop ();
    }

    pthread_mutex_lock (&track_mutex);
    old_current_track = current_track;
    //    next_track ();
    current_track++;
    log_debug (DEBUG_CDPLAYER, "next_track: current_track= %d", current_track);
    pthread_mutex_unlock (&track_mutex);

    if (old_mode == MODE_PLAYING)
      play_track (current_track);
  }
}

static void mp3_player_previous_track ()
{
  int old_mode = mp3_mode;

  if (current_track > 1) {
    if (mp3_mode != MODE_STOPPED) {
      if (mp3_mode == MODE_PAUSED)
	mp3_end = 1;
      do_stop ();
    }

    pthread_mutex_lock (&track_mutex);
    old_current_track = current_track;
    //    previous_track ();
    current_track--;
    pthread_mutex_unlock (&track_mutex);

    if (old_mode == MODE_PLAYING)
      play_track (current_track);
  }
}

/* Toggle time+track/title display */
static void mp3_player_toggle_display ()
{
  /* Next type */
  display_type++;

  if (display_type == DISPLAY_LAST_TYPE || (!tracks[current_track-1]->tag && display_type == DISPLAY_LAST_TYPE_NOID3))
    display_type = 0;

  switch (display_type) {
  case DISPLAY_TOTAL_LENGTH:
    display_total_length ();
    break;

  case DISPLAY_TRACK_LENGTH:
    display_track_length ();
    break;

  default:
    lcd_init_hscroll ();
    break;
  }
}

static void mp3_player_reload ()
{
  char buf[MAXPATHLEN], c = 0;
  int i = 0;

  log_debug (DEBUG_MP3PLAYER, "mp3player_reload: starting");

  /* Stop if playing */
  mp3_player_stop ();
  
  /* Wait for real stop */
  while (mp3_mode != MODE_STOPPED)
    usleep (100000);

  /* This should force info process to exit */
  reloading = TRUE;

  /* Wait for info process to exit */
  while (info_process_running)
    usleep (100000);

  /* Free the tracks */
  if (max_track > 0)
    free_tracks (0, max_track-1);

  /* Now load the playlists */
  max_track = 0;
  lcd_write_center (1, "Loading playlists", ATTR_NORMAL);
  while (read (fd_fifo, &c, 1) > 0) {

    /* End is EOT */
    if (c == 0x04)
      break;

    if (c == '\n') {
      buf[i] = 0;

      /* Load the playlist */
      log_debug (DEBUG_MP3PLAYER, "Loading playlist %s from index %d", buf, max_track);
      max_track += load_play_list (buf, playlist, max_track);
      i = 0;
    }
    else
      buf[i++] = c;
  }


  /* Check for correct loading */
  if (max_track <= 0) 
    lcd_write_center (1, "No track to play !", ATTR_NORMAL);
  else {
    reloading = FALSE;

    /* Start time displaying thread */
    if (pthread_create (&display_info_thread, &thread_attr, display_info_process, NULL) != 0)
      log_err ("pthread_create error for display_info_process");
  }

  log_debug (DEBUG_MP3PLAYER, "mp3player_reload: exiting");
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s [-s] [-d playlist dir] [-p list of playlists] [list of files]\n", s);
  exit (1);
}

/* Main function */
main (int ac, char **av)
{
  char replybuf[256], *cp;
  int i = 0;

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if (ac == 1)
    usage (progname);

  init_debug ();
  log_debug (DEBUG_MP3PLAYER, "starting.");

  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {
      case 'p':
	playlist_list = ++av; break;

      case 'd':
	playlist_dir = *++av; break;

      case 's':
	use_stdin = TRUE; break;

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }


  if (pipe (pipe_in) < 0) {
    log_err ("pipe_in: %s", strerror(errno));
    exit (1);
  }

  if (pipe (pipe_out) < 0) {
    log_err ("pipe_out: %s", strerror(errno));
    exit (1);
  }

  if (!(mpg123_pid = fork ())) {
    dup2 (pipe_out[0], 0);
    close (pipe_out[0]);
    dup2 (pipe_in[1], 1);
    close (pipe_in[1]);
    close (2);

    if (execlp (MPG123, MPG123, "-R", "-", NULL) < 0) {
      log_err ("execlp: %s: %s", MPG123, strerror(errno));
      exit (1);
    }
  }
  else {
    signal (SIGINT, got_interrupt);
    signal (SIGQUIT, got_interrupt);
    signal (SIGTERM, got_interrupt);

    read_mpg123_status (replybuf, sizeof(replybuf));
    log_debug (DEBUG_MP3PLAYER, "mpg123 version= %s\n", replybuf);

    /* Init API */
    omm_api_init (mp3_event_handlers, MP3PLAYER_FIFO, use_stdin);
    lcd_clear_screen ();

    /* Load list from argv or file */
    if (!playlist_list) {
      register int i = 0;

      lcd_write_center (1, "Loading tracks", ATTR_NORMAL);
      while (*av && i < MAXTRACKS-1)
	playlist[i++] = strdup (*av++);

      max_track = load_files (playlist, 0);
    }
    else {
      lcd_write_center (1, "Loading playlists", ATTR_NORMAL);
      for (i = 0 ; playlist_list[i] ; i++) {
	log_debug (DEBUG_MP3PLAYER, "Loading playlist %s from index %d", playlist_list[i], max_track);
	max_track += load_play_list (playlist_list[i], playlist, max_track);
      }
    }

    if (max_track <= 0) 
      lcd_write_center (1, "No track to play !", ATTR_NORMAL);
    else {
      /* Start time displaying thread */
      if (pthread_create (&display_info_thread, &thread_attr, display_info_process, NULL) != 0)
	log_err ("pthread_create error for display_info_process");
    }

    pthread_mutex_init (&mpg123_mutex, NULL);
    pthread_mutex_init (&track_mutex, NULL);

    /* Main loop */
    omm_api_main_loop ();

    /* End */
    log_debug (DEBUG_MP3PLAYER, "exiting.");
    pthread_mutex_destroy (&mpg123_mutex);
    pthread_mutex_destroy (&track_mutex);
    close_debug ();
  }
}
