#ident "$Id$"
/*
 * OMM Napster client
 */

#define EXTERN
#include "omm.h"
#include "napster.h"

static char *omm_dir = DEFAULT_OMMDIR;
static int nap_pid;
static int pipe_out[2];
static int fd_napio;
static pthread_t read_nap_thread;
static pthread_t scrolled_thread;
static pthread_t download_thread;
static char *scrolled_str;
static char song_name[1024], old_song_name[1024];
static pthread_mutex_t scroll_mutex;
static pthread_mutex_t down_mutex;
static pthread_mutex_t line0_mutex;
static int results;
static int idown;
static boolean title_cleared;

/* States handling definition */
static omm_event_handler napster_event_handlers[] = {
  'q', OMM_STATE_INIT, napster_quit,		/* Quit */
  'q', STATE_CONNECTED, napster_quit,
  'q', STATE_RESULTS, napster_quit,
  'q', STATE_DOWNLOAD, napster_quit,
  'S', STATE_CONNECTED, napster_search,         /* Search */
  'S', STATE_RESULTS, napster_search,
  'S', STATE_DOWNLOAD, napster_search,
  'j', STATE_RESULTS, napster_next_song,        /* Next/prev song */
  'k', STATE_RESULTS, napster_prev_song,
  'j', STATE_DOWNLOAD, napster_next_download,
  'k', STATE_DOWNLOAD, napster_prev_download,
  '\n', STATE_RESULTS, napster_download,         /* Download current song */
  'd', STATE_RESULTS, napster_update,           /* Back to current download */
  'd', STATE_DOWNLOAD, napster_update,          /* Update current download */
  'r', STATE_DOWNLOAD, napster_results,         /* Back to results */
  'D', STATE_DOWNLOAD, napster_delete,          /* Delete current download */
  'B', OMM_STATE_INIT, lcd_bg_screen,           /* Appli in background */
  'B', STATE_CONNECTED, lcd_bg_screen,
  'B', STATE_RESULTS, lcd_bg_screen,
  'B', STATE_DOWNLOAD, lcd_bg_screen,
  'F', OMM_STATE_INIT, lcd_fg_screen,           /* Appli in foreground */
  'F', STATE_CONNECTED, lcd_fg_screen,
  'F', STATE_RESULTS, lcd_fg_screen,
  'F', STATE_DOWNLOAD, lcd_fg_screen,
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, napster_default_handler
};

/* 
 * Set the scrolled string 
 * scroll_mutex should be locked before entering
 */
static void set_scrolled_string ()
{
  if (strcmp (song_name, old_song_name)) {
    /* The string has changed */
    log_debug (DEBUG_NAPSTER, "set_scrolled_string: changed!");
    strncpy (old_song_name, song_name, sizeof(old_song_name));

    /* Clear line 1 and re-init scrolling */
    lcd_clear_line (1);
    lcd_init_hscroll ();
    scrolled_str = song_name;
  }
}

/* Set the program state */
static void napster_set_state (int state)
{
  log_debug (DEBUG_NAPSTER, "napster_set_state: %d --> %d", omm_state, state);
  omm_state = state;
}

static void got_interrupt (int sig)
{	
  log_debug (DEBUG_NAPSTER, "Got interrupt %d!", sig);

  napster_quit ();
  exit (1);
}

/* Write a warning string (line 0) */
static void warn_string (char *s, int t)
{
  pthread_mutex_lock (&line0_mutex);
  lcd_save_area (0, 0, LCD_WIDTH-1, 0);
  lcd_write_center (0, s, ATTR_BLINK);
  sleep (t);
  lcd_restore_area (0, 0, LCD_WIDTH-1, 0);
  pthread_mutex_unlock (&line0_mutex);
}

/* Send a string to the external Napster client */
static void send_nap (char *s)
{
  log_debug (DEBUG_NAPSTER, "send_nap: %s", s);
  write (NAP_OUT, s, strlen(s));
}

/* Display the current scrolled string (line 1) */
static void *display_scrolled_process (void *dummy)
{
  for (;;) {
    pthread_mutex_lock (&scroll_mutex);

    if (scrolled_str)
      lcd_write_hscroll (-1, 1, scrolled_str);

    pthread_mutex_unlock (&scroll_mutex);
    usleep (500000);
  }
}

/* 
 * Update the current download window: just get the last @D line
 * from the external Napster client.
 */
static void *display_download_process (void *dummy)
{
  for (;;) {
    pthread_mutex_lock (&down_mutex);
    if (omm_state == STATE_DOWNLOAD)
      send_nap ("/updown\n");
    pthread_mutex_unlock (&down_mutex);
    sleep (1);
  }
}


/* Next/prev song in STATE_RESULTS state */
static void napster_next_song ()
{
  send_nap ("j");
}

static void napster_prev_song ()
{
  send_nap ("k");
}

/*
 * Start downloading the current song: just send LF then 'q' and /pdown to
 * get the first @D line.
 */
static void napster_download ()
{
  pthread_mutex_lock (&down_mutex);
  napster_set_state (STATE_DOWNLOAD);
  send_nap ("\nq");
  sleep (1);
  send_nap ("/pdown\n");
  pthread_mutex_unlock (&down_mutex);
}

/* Delete current download */
static void napster_delete ()
{
  char buf[256];

  sprintf (buf, "/ddown %d\n", idown);
  send_nap (buf);
}

/*
 * Update the download info: get the last @D line
 */
static void napster_update ()
{
  /* If STATE_RESULTS send 'q' to change the window */
  if (omm_state == STATE_RESULTS)
    send_nap ("q");

  pthread_mutex_lock (&down_mutex);
  napster_set_state (STATE_DOWNLOAD);
  send_nap ("/updown\n");
  pthread_mutex_unlock (&down_mutex);

}

/* Display next/prev download */
static void napster_next_download ()
{
  pthread_mutex_lock (&down_mutex);
  send_nap ("/npdown\n");
  pthread_mutex_unlock (&down_mutex);
}

static void napster_prev_download ()
{
  pthread_mutex_lock (&down_mutex);
  send_nap ("/ppdown\n");
  pthread_mutex_unlock (&down_mutex);
}

/* Quit */
static void napster_quit ()
{
  if (nap_pid)
    kill (nap_pid, SIGTERM);
  close (fd_napio);

  napster_set_state (OMM_STATE_INIT);
  omm_api_close ();
}

/* Input string validation */
void input_ok ()
{
  char buf[256], *s;

  /* Get the string */
  s = omm_api_get_input_str();

  /* Send the search command to the external Napster client */
  sprintf (buf, "/search %s\n", s);
  //  sprintf (buf, "/search madonna\n", s);
  send_nap (buf);

  /* So we go to STATE_RESULTS */
  napster_set_state (STATE_RESULTS);
}

/* Abort input */
void input_aborted ()
{
  lcd_restore_area (0, 1, LCD_WIDTH-1, 3);
  log_debug (DEBUG_NAPSTER, "input ABORTED!");
  /* Now we are back to OMM_STATE_INIT */
}

/* Search a string */
static void napster_search ()
{
  lcd_save_area (0, 1, LCD_WIDTH-1, 3);
  omm_api_start_input ("Search ?", input_ok, input_aborted);
}

/* Go to STATE_RESULTS state */
static void napster_results ()
{
  if (results) {
    napster_set_state (STATE_RESULTS);
    send_nap ("/results\n");
  }
  else
    log_debug (DEBUG_NAPSTER, "napster_results: no results!");
}

/* Read a \n terminated string from the external Napster client */
static void read_nap_fifo (char *s, int size)
{
  char c = 0;
  int cnt = 0;

  memset ((void*)s, 0, size);
  while (c != '\n' && cnt < size) {
    if (read (fd_napio, &c, 1) < 0) {
      log_err ("read_nap_fifo: %s: %s", s, strerror(errno));
      napster_quit ();
    }

    if (c != '\n') {
      *(s+cnt) = c;
      cnt++;
    }
  }
}

/* 
 * Read status lines from external Napster client:
 *
 * @C        connexion status
 * @M        message
 * @M        song description
 * @D        download status
 * @F        download finished
 */
static void *read_nap_process (void *dummy)
{
  char buf[256], user[256], conntype[256];
  int percent, idownf, ndown, size, brate, y;
  float rate;
  
  log_debug (DEBUG_NAPSTER, "read_nap_process: starting");

  while (1) {
    read_nap_fifo (buf, sizeof(buf));

    //    log_debug (DEBUG_NAPSTER, "read_nap_process: buf= '%s'", buf);

    switch (buf[1]) {

      /* Connexion */
    case 'C':
      /* Connected to Napster server */
      if (buf[3] == '1') {
	napster_set_state (STATE_CONNECTED);
	lcd_write_center (2, "Connected!", ATTR_NORMAL);
      }
      else
	lcd_write_center (2, "Connection error!", ATTR_NORMAL);
      break;

    case 'I':
      /* Delete download */
      idownf = atoi(&buf[3]);
      /* Write a blinking message on line 0 */
      sprintf (buf, "DELETED %d", idownf);
      warn_string (buf, 3);
      break;

      /* Message */
    case 'M':
      /* Format is '@M y message' */
      y = buf[3] - '0';

      if (y == 0)
	pthread_mutex_lock (&line0_mutex);

      /* Check if space for 'NAP' + message on line 0 */
      if (y == 0 && strlen (&buf[5]) < LCD_WIDTH-4) {
	if (title_cleared) {
	  /* Redraw title */
	  lcd_write (0, y, "NAP", ATTR_NORMAL);
	  title_cleared = 0;
	}
	lcd_write (4, y, &buf[5], ATTR_NORMAL);
      }
      else {
	if (y == 0)
	  title_cleared = 1;

	lcd_write_center (y, &buf[5], ATTR_NORMAL);
      }

      if (y == 0)
	pthread_mutex_unlock (&line0_mutex);

      break;
      
      /* Song */
    case 'S':
      /* Got a result */
      results++;

      size = atoi(strtok (&buf[3], "|"));
      brate = atoi(strtok (NULL, "|"));
      strncpy (conntype, strtok (NULL, "|"), sizeof(conntype));

      pthread_mutex_lock (&scroll_mutex);
      strncpy (song_name, strtok (NULL, "|"), sizeof(song_name));
      set_scrolled_string ();
      pthread_mutex_unlock (&scroll_mutex);

      sprintf (buf, "%d bytes", size);
      lcd_write_center (2, buf, ATTR_NORMAL);

      sprintf (buf, "%dK %s", brate, conntype);
      lcd_write_center (3, buf, ATTR_NORMAL);

      log_debug (DEBUG_NAPSTER, "size= %d bytes rate= %dK conn= %s song= '%s'", size, brate, conntype, song_name);

      break;

    case 'D': /* Download */
      switch (buf[3]) {

	/* No more download */
      case 'N':
	log_debug (DEBUG_NAPSTER, "read_nap_process: no download!");
	lcd_clear_line (1);
	lcd_clear_line (2);
	lcd_clear_line (3);
	lcd_write_center (2, "No download!", ATTR_NORMAL);

	/* Back to STATE_RESULTS */
	napster_results ();
	break;

	/* Waiting */
      case 'W':
	idown = atoi(strtok (&buf[5], "|"));
	ndown = atoi(strtok (NULL, "|"));
	strncpy (user, strtok (NULL, "|"), sizeof(user));

	pthread_mutex_lock (&scroll_mutex);
	strncpy (song_name, strtok (NULL, "|"), sizeof(song_name));
	set_scrolled_string ();
	pthread_mutex_unlock (&scroll_mutex);

	sprintf (buf, "%d/%d waiting from", idown, ndown);
	lcd_write_center (2, buf, ATTR_NORMAL);
	lcd_write_center (3, user, ATTR_NORMAL);

	log_debug (DEBUG_NAPSTER, "read_nap_process: %d waiting user= %s song='%s'", ndown, user, song_name);
	break;

	/* Getting */
      case 'G':
	idown = atoi(strtok (&buf[5], "|"));
	ndown = atoi(strtok (NULL, "|"));
	rate = 0.00001 + atof(strtok (NULL, "|"));
	percent = atoi(strtok (NULL, "|"));
	strncpy (user, strtok (NULL, "|"), sizeof(user));

	pthread_mutex_lock (&scroll_mutex);
	strncpy (song_name, strtok (NULL, "|"), sizeof(song_name));
	set_scrolled_string ();
	pthread_mutex_unlock (&scroll_mutex);

	sprintf (buf, "%d/%d %.1fK/s %d%%", idown, ndown, rate, percent);
	lcd_write_center (2, buf, ATTR_NORMAL);
	lcd_write_center (3, user, ATTR_NORMAL);

	log_debug (DEBUG_NAPSTER, "read_nap_process: %d downloading %.1f K/s %d%% user= %s song='%s'", ndown, rate, percent, user, song_name);
	break;

      default: break;
      }

      break;

    case 'F': /* download Finished */
      idownf = atoi(&buf[3]);
      /* Write a blinking message on line 0 */
      sprintf (buf, "FINISHED %d", idownf);
      warn_string (buf, 3);
      break;
      
    default:
      log_debug (DEBUG_NAPSTER, "read_nap_process: %s", buf);
      break;
    }
  }
}


/* Default input handler */
static void napster_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  log_debug (DEBUG_NAPSTER, "napster_default_handler: got '%c' in state %02x", ev, omm_state);
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s\n", s);
  exit (1);
}

/* Main function */
main (int ac, char **av)
{
  char *cp, *new_omm_dir, buf[MAXPATHLEN];

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if ((new_omm_dir = getenv ("OMMDIR")))
    omm_dir = new_omm_dir;

  init_debug ();

  /* Get options */
  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }

  log_debug (DEBUG_NAPSTER, "starting.");

  if (pipe (pipe_out) < 0) {
    log_err ("pipe_out: %s", strerror(errno));
    exit (1);
  }

  if ((fd_napio = open_fifo (NAPIO)) < 0)
    exit (1);

  /* Init API */
  omm_api_init (napster_event_handlers, NAPSTER_FIFO, FALSE);

  lcd_clear_screen ();
  lcd_write (0, 0, "NAP", ATTR_NORMAL);
  pthread_mutex_init (&line0_mutex, NULL);

  /* Create str display process */
  pthread_mutex_init (&scroll_mutex, NULL);
  if (pthread_create (&scrolled_thread, &thread_attr, display_scrolled_process, NULL) != 0)
    log_err ("pthread_create error for display_scrolled_process");

  pthread_mutex_init (&down_mutex, NULL);
  if (pthread_create (&download_thread, &thread_attr, display_download_process, NULL) != 0)
    log_err ("pthread_create error for display_download_process");


  if (pthread_create (&read_nap_thread, &thread_attr, read_nap_process, NULL) != 0)
    log_err ("pthread_create error for read_nap_process");

  if (!(nap_pid = fork ())) {
    dup2 (pipe_out[0], 0);
    close (pipe_out[0]);

    putenv ("LINES=4");

    sprintf (buf, "%s/conf/nap.conf", omm_dir);

    //    if (execlp (NAP, NAP, "-f", buf, "-t", "-b", "-o", NAPIO, NULL) < 0) {
    if (execlp (NAP, NAP, "-f", buf, "-t", "-o", NAPIO, NULL) < 0) {
      log_err ("execlp: %s: %s", NAP, strerror(errno));
      exit (1);
    }
  }
  else {
    signal (SIGINT, got_interrupt);
    signal (SIGQUIT, got_interrupt);
    signal (SIGTERM, got_interrupt);

    /* Main loop */
    omm_api_main_loop ();

    /* End */
    log_debug (DEBUG_NAPSTER, "exiting.");
    close_debug ();
  }
}

