/*
** Antoine Terriene <antoine.terrienne@free.fr>
**
** Latency bench for rtai
** based on Andrew Morton's amlat
**
** in-kernel version (user side part)
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

#define DEF_HIST_FILE	"rtai_amlat.hist"

#define SAMPLES	10000000
int histogram[SAMPLES + 1];		/* Milliseconds */
char *file;

void hist(void)
{
  int i;
  FILE *f;
  
  f = fopen(file, "w");
  if (f == 0)
    {
      fprintf(stderr, "realfeel: can't open `%s':%s\n",
	      file, strerror(errno));
      exit(1);
    }
  
  for (i = 0; i < SAMPLES; i++)
    {
      if (histogram[i])
	{
	  fprintf(f, "%d %d\n",
		  i, histogram[i]);
	}
    }
  fclose(f);
  exit(0);
}

void	int_handle(int a)
{
  hist();
}


int	main(int argc, char **argv)
{
  int		fifo;
  double	ideal;
  
  signal(SIGINT, int_handle);
  file = DEF_HIST_FILE;
  if (argc > 1)
    {
      file = argv[1];
    }
  
  memset(histogram, 0, SAMPLES * sizeof (int));
  
  if ((fifo = open("/dev/rtf0", O_RDONLY)) < 0)
    {
      perror("Error opening /dev/rtf0 ");
      return(1);
    }
  
  ideal = 1.0 / 2048/*hz*/;
  
  for (;;)
    {
      char	cmd;
      long long	delay;
      int	ms;
      
      if (read(fifo, &cmd, 1) != 1)
	{
	  perror("cannot read fifo ");
	  return (1);
	}
      
      if (cmd == 'Q')
	{
	  printf("leaving rtai_amlat\n");
	  hist();
	  return (0);
	}
      
      if (read(fifo, &delay, sizeof (delay)) == -1)
	{
	  perror("cannot read fifo ");
	  return (1);
	}
      
      ms = delay / 1000;
      if (ms < 0)
	{
	  ms = 0;		/* hmmm */
	}
      if (ms >= SAMPLES)
	{
	  ms = SAMPLES;
	}
      histogram[ms]++;
      
    }
  
  return (0);
}
