/*
** Antoine Terrienne <antoine.terrienne@free.fr>
**
** Latency bench for rtai
** based on Andrew Morton's amlat
**
** lxrt version
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>
#include <sys/mman.h>
#include <math.h>
#include <signal.h>

#define KEEP_STATIC_INLINE
#include <rtai_lxrt_user.h>
#include <rtai_lxrt.h>

/* 2Hz 488us 4882812ns */
#define TICK_PERIOD period
#define TASK_PRIORITY 1
#define STACK_SIZE 10000

static long	period = 488281;
static long	nb;

static RT_TASK	*rt_task;
static long	nb_count = 2048 * 60;

#define DEF_HIST_FILE	"rtai_amlat.hist"

#define SAMPLES	10000000
int histogram[SAMPLES + 1];		/* Milliseconds */
char *file;

void hist(void)
{
  int i;
  FILE *f;
  
  f = fopen(file, "w");
  if (f == 0) {
    fprintf(stderr, "realfeel: can't open `%s':%s\n",
	    file, strerror(errno));
    exit(1);
  }
  
  for (i = 0; i < SAMPLES; i++) {
    if (histogram[i]) {
      fprintf(f, "%d %d\n",
	      i, histogram[i]);
    }
  }
  fclose(f);
  exit(0);
}

void	int_handle(int a)
{
  hist();
}

RTIME		tick_period;

RTIME	last;

static void	local_main(void)
{
  double	ideal;
  RTIME		now;
  RTIME		delta = 0;
  
  while (nb_count)
    {
      int	ms;
      
      rt_task_wait_period();
      now = rt_get_time();
      delta = count2nano((now - last) - tick_period);
      ms = delta / 1000;
      if (ms < 0)
	{
	  ms = 0;		/* hmmm */
	}
      if (ms >= SAMPLES)
	{
	  ms = SAMPLES;
	}
      histogram[ms]++;
      nb_count--;
      last = now;
    }
}

int		main(int argc, char **argv)
{
  int		i;
  unsigned long	mytask_name;
  
  printf("Rtai amlat <antoine.terrienne@free.fr>\n");
  
  signal(SIGINT, int_handle);
  file = DEF_HIST_FILE;
  if (argc < 3)
    {
      printf("Usage : %s nsec_period nb_samples [out_file]\n", argv[0]);
      return (1);
    }
  
  period = atoi(argv[1]);
  nb_count = atoi(argv[2]);
  if (argc > 3)
    {
      file = argv[3];
    }
  
  memset(histogram, 0, SAMPLES * sizeof (int));
  
  /* rt_set_periodic_mode(); */
  rt_set_oneshot_mode();
  mytask_name = nam2num("AMLAT_LXRT");
  mlockall(MCL_CURRENT | MCL_FUTURE);
  
  if (!(rt_task = rt_task_init(mytask_name, 1, 0, 0)))
    {
      printf("CANNOT INIT TASK %lu\n", mytask_name);
      return (1);
    }
  
  tick_period = start_rt_timer(nano2count(TICK_PERIOD));
  printf("period : %lli\n", count2nano(tick_period));
  last = rt_get_time();
  if (i = rt_task_make_periodic(rt_task, rt_get_time() /* + tick_period */,
				tick_period))
    {
      rt_printk("error at starting task : %i\n", i);
    }
  
  rt_make_hard_real_time();
  local_main();
  rt_make_soft_real_time();
  
  stop_rt_timer();
  rt_task_delete(rt_task);
  hist();
  return (1);
}
