/*
** Antoine Terrienne <antoine.terrienne@free.fr>
**
** Latency bench for rtai
** based on Andrew Morton's amlat
**
** in-kernel version
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>
#include <math.h>
#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>

/* 2Hz 488us 488281ns */
#define TICK_PERIOD period /*488281*/
#define TASK_PRIORITY 1
#define STACK_SIZE 10000
#define FIFO 0

MODULE_LICENSE("GPL");

long	period = 488281; /* 2048Hz */
RTIME	tick_period;
long	nbsamples = 122880; /* for 1 min (2048 * 60)*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(period, "1-1000000000l");
MODULE_PARM(nbsamples, "1-1000000000l");
#endif

static RT_TASK	rt_task;
/* static long	nb_count = 2048 * 60; */
#define nb_count	nbsamples

RTIME	last;

static void	local_main(int a)
{
  RTIME		now;
  RTIME		delta = 0;
  
  /* rt_printk("first time : %lli\n", rt_get_time_ns()); */
  while (nb_count)
    {
      rt_task_wait_period();
      now = rt_get_time();
      delta = count2nano((now - last) - tick_period);
      /* rt_printk("last %lli now %lli delta %lli\n",
		count2nano(last),
		count2nano(now), delta); */
      rtf_put(FIFO, "T", 1);
      rtf_put(FIFO, &delta, sizeof (delta));
      nb_count--;
      last = now;
    }
  /* send stop signal */
  rtf_put(FIFO, "Q", 1);
}

int	init_module(void)
{
  int	i;
  
  rt_printk("Rtai amlat <antoine.terrienne@free.fr>\n");
  /* rt_set_periodic_mode(); */
  rt_set_oneshot_mode();
  if (i = rt_task_init(&rt_task, local_main, 1, STACK_SIZE, TASK_PRIORITY,
		       1, 0))
    {
      rt_printk("error at creating task : %i\n", i);
    }
  rtf_create(FIFO, nb_count * (sizeof (long) + 1));
  
  tick_period = start_rt_timer(nano2count(TICK_PERIOD));
  
  last = rt_get_time();
  rt_printk("period : %lli\n", count2nano(tick_period));
  /* rt_printk("first time : %lli\n", count2nano(last)); */
  if (i = rt_task_make_periodic(&rt_task, rt_get_time() /* + tick_period */,
				tick_period))
    {
      rt_printk("error at starting task : %i\n", i);
    }
  
  return 0;
}

void	cleanup_module(void)
{
  stop_rt_timer();
  rtf_destroy(FIFO);
  rt_task_delete(&rt_task);
}
