/*
** Antoine Terrienne <antoine.terrienne@free.fr>
**
** Latency bench for rtai
** based on Andrew Morton's amlat
**
** in-kernel version
*/

/* #include <linux/kernel.h>
   #include <linux/module.h> */
/*#include <asm/io.h>
  #include <math.h>*/
#include <rtl.h>
/* #include <linux/time.h> */
#include <rtl_time.h>
#include <pthread.h>
#include <rtl_fifo.h>

/* 2Hz 488us 488281ns */
#define TICK_PERIOD period /*488281*/
#define TASK_PRIORITY 1
#define STACK_SIZE 10000
#define FIFO 0

long		period = 488281; /* 2048Hz */
long		nbsamples = 122880; /* for 1 min (2048 * 60)*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(period, "1-1000000000l");
MODULE_PARM(nbsamples, "1-1000000000l");
#endif

static pthread_t	thread;
/* static long	nb_count = 2048 * 60; */
#define nb_count	nbsamples


static void		*local_main(void *a)
{
  hrtime_t		tick_period;
  hrtime_t		last;
  hrtime_t		now;
  hrtime_t		delta;
  struct sched_param	p;
  
  tick_period = period;
  p.sched_priority = 1;
  pthread_setschedparam (pthread_self(), SCHED_FIFO, &p);
  
  last = gethrtime();
  pthread_make_periodic_np (pthread_self(), last, period);
  
  while (nb_count)
    {
      pthread_wait_np();
      now = gethrtime();
      delta = (now - last) - tick_period;
      rtf_put(FIFO, "T", 1);
      rtf_put(FIFO, &delta, sizeof (delta));
      nb_count--;
      last = now;
    }
  rtf_put(FIFO, "Q", 1);
  return (NULL);
}

int	init_module(void)
{
  rtf_create(FIFO, nb_count * (sizeof (long) + 1));
  rtl_printf("Rtai amlat <antoine.terrienne@free.fr>\n");
  return (pthread_create(&thread, NULL, local_main, 0));
}

void	cleanup_module(void)
{
  pthread_cancel(thread);
  rtf_destroy(FIFO);
  pthread_join(thread, NULL);
}
