#!/bin/sh

TMP_DIR=/tmp
TMP_HD=/dev/hda

if [ $# -ne 1 ]
then
    echo "Usage : testing.sh <tick_hz> <leng_sec> [test_level]"
    echo "test_level = 0 no charge"
    echo "test_level = 1 soft charge (burnMMX, crashme)"
    echo "test_level = 2 hd charge (dd)"
    echo "test_level = 3 soft + hd charge"
    echo "no test_level, assuming test_level = 0"
fi

USEC_PERIOD=$((1000000 / $1))
NSEC_PERIOD=$((1000000000 / $1))

NB_TICKS=$(( $2 * $1 ))

test_level=$3
type=`uname -r`
htmlfile=$type-$1-$2.html
PATH=$PATH:/home/antoine/rt/bench_tools/ctcs-1.3.0pre4/runin/bin

hash -r
hdparm -d 1 $TMP_HD

case $type in
    2.4*-rthal*)
    echo rtai detected
    ;;
    2.4*-rtl)
    echo rtLinux detected
    ;;
    2.4.*)
    echo regular linux detected
    ;;
    *)
    echo Cannot define running system.
    exit
    ;;
esac

touch $TMP_DIR/amlat_loop

case $test_level in
    1)
        crashme +1024 `random` 0 &
	burnMMX &
    ;;
    2)
        while [ -f $TMP_DIR/amlat_loop ]; do dd if=/dev/zero of=$TMP_DIR/my_amlat bs=100k count=100; sync; done &
    ;;
    3)
        crashme +1024 `random` 0 &
	burnMMX &
	while [ -f $TMP_DIR/amlat_loop ]; do dd if=/dev/zero of=$TMP_DIR/my_amlat bs=100k count=100; sync; done &
    ;;
    *)
    ;;
esac

sleep 1

case $type in
    2.4.*-rthal*)
        # testing in-kernel rtai
        insmod rtai
        insmod rt_mem_mgr
	insmod rtai_fifos
	insmod rtai_sched
	insmod rtai_amlat_module.o period=$NSEC_PERIOD nbsamples=$NB_TICKS
	./rtai_amlat rtai_kernel_latency_$1-$2-$3.hist
	rmmod rtai_amlat_module
        # testing lxrt rtai
	insmod rtai_lxrt
	./rtai_amlat_lxrt $NSEC_PERIOD $NB_TICKS rtai_lxrt_latency_$1-$2-$3.hist
	rmmod rtai_lxrt
	rmmod  rtai_sched rtai_fifos rt_mem_mgr rtai
        # testing user space rtai
	insmod rtc
	./linux_amlat $1 $NB_TICKS rtai_user_latency_$1-$2-$3.hist
	rmmod rtc
    ;;
    2.4.*-rtl)
        insmod rtlinux_amlat_module.o period=$NSEC_PERIOD nbsamples=$NB_TICKS
	./rtai_amlat rtlinux_kernel_latency_$1-$2-$3.hist
	rmmod rtlinux_amlat_module
        # testing user space rtLinux
	insmod rtc
	./linux_amlat $1 $NB_TICKS rtlinux_user_latency_$1-$2-$3.hist
	rmmod rtc
    ;;
    2.4.*)
        insmod rtc
	./linux_amlat $1 $NB_TICKS linux_$type-$1-$2-$3.hist
	rmmod rtc
    ;;
esac

case $test_level in
    1)
        killall crashme
        killall burnMMX
    ;;
    2)
        rm $TMP_DIR/amlat_loop
    ;;
    3)
        killall crashme
        killall burnMMX
	rm $TMP_DIR/amlat_loop
    ;;
    *)
    ;;
esac

