/*
COPYRIGHT (C) 2001  Paolo Mantegazza (mantegazza@aero.polimi.it),

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <signal.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <asm/rtai_srq.h>
#include <rtai_shm.h>

#include "kurtnet.h"

#define ADRSZ  sizeof(struct sockaddr_in)

static struct sock_t *runsock, *socks;

static int srq;

static void *recv_fun(void *thread_arg)
{
	struct sched_param mysched;
	fd_set sfd_list, w_sfd_list;
	int i, arg[2] = { RECVFROM, 0 };

	FD_ZERO(&sfd_list);
	for (i = 0; i < (MAX_STUBS + MAX_SOCKS); i++) {
		socks[i].sock = socket(AF_INET, SOCK_DGRAM, 0);
		fcntl(socks[i].sock, O_NONBLOCK);
		bind(socks[i].sock, &socks[i].bindaddr, ADRSZ);
		FD_SET(socks[i].sock, &sfd_list);
	}
        mysched.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if (sched_setscheduler(0, SCHED_FIFO, &mysched) == -1) {
		puts("ERROR IN SETTING URTNET RECV SERV TO SCHED_FIFO\n");
        	return (void *)1;
 	}       
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (1) {
		w_sfd_list = sfd_list;
		select(socks[MAX_STUBS + MAX_SOCKS - 1].sock + 1, &w_sfd_list, NULL, NULL, NULL);
		for (i = 0; i < (MAX_STUBS + MAX_SOCKS); i++) {
			if (FD_ISSET(socks[i].sock, &w_sfd_list)) {
				socks[i].addrlen = ADRSZ;
				socks[i].recv = recvfrom(socks[i].sock, socks[i].msg, MAX_MSG_SIZE, 0, &socks[i].addr, &socks[i].addrlen);
				arg[1] = i;
				rtai_srq(srq, (unsigned int)arg);
			}
		}
	}
        return (void *)0;
}

void endme (int dummy)
{
	signal(SIGTERM, endme);
}

static pthread_t recv_thread;

int main(void)
{
	int i, arg[2] = { SENDTO, 0 };
	struct sched_param mysched;

	signal(SIGTERM, endme);

	srq = rtai_open_srq(0xcacca1);
        mysched.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if (sched_setscheduler(0, SCHED_FIFO, &mysched) == -1) {
		puts("ERROR IN SETTING URTNET MAIN TO SCHED_FIFO\n");
		exit(1);
 	}       
        runsock = (struct sock_t *)rtai_malloc(0xcacca0, (MAX_STUBS + MAX_SOCKS + 1)*sizeof(struct sock_t));
        socks = runsock + 1;
        runsock->opnd = 1;
	if (pthread_create(&recv_thread, NULL, recv_fun, NULL)) {
		printf("ERROR IN CREATING RECV SERVER\n");
                exit(1);
        }
	mlockall(MCL_CURRENT | MCL_FUTURE);

	while (runsock->opnd) {
		if ((i = rtai_srq(srq, (unsigned int)arg)) >= 0) {
			sendto(socks[i].sock, socks[i].msg, socks[i].len, 0, &socks[i].addr, ADRSZ);
		}
	}
	for (i = 0; i < (MAX_STUBS + MAX_SOCKS); i++) {
		close(socks[i].sock);
	}
	rtai_free(0xcacca0, socks);
	exit(0);
}
